import pytest

from xdsl.dialects.builtin import IntegerAttr, StringAttr
from xdsl.dialects.smt import (
    AndOp,
    ApplyFuncOp,
    AssertOp,
    BinaryBVOp,
    BitVectorAttr,
    BitVectorType,
    BoolType,
    BVAddOp,
    BVAndOp,
    BVAShrOp,
    BvConstantOp,
    BVLShrOp,
    BVMulOp,
    BVNegOp,
    BVNotOp,
    BVOrOp,
    BVSDivOp,
    BVShlOp,
    BVSModOp,
    BVSRemOp,
    BVUDivOp,
    BVURemOp,
    BVXOrOp,
    ConstantBoolOp,
    DeclareFunOp,
    DistinctOp,
    EqOp,
    ExistsOp,
    ForallOp,
    FuncType,
    ImpliesOp,
    IteOp,
    NotOp,
    OrOp,
    QuantifierOp,
    UnaryBVOp,
    VariadicBoolOp,
    XOrOp,
    YieldOp,
)
from xdsl.ir import Block, Region
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import create_ssa_value


def test_constant_bool():
    op = ConstantBoolOp(True)
    assert op.value is True
    assert op.value_attr == IntegerAttr(-1, 1)

    op = ConstantBoolOp(False)
    assert op.value is False
    assert op.value_attr == IntegerAttr(0, 1)


def test_bv_type():
    bv_type = BitVectorType(32)
    assert bv_type.width.data == 32
    assert bv_type.value_range() == (0, 2**32)

    bv_type = BitVectorType(3)
    assert bv_type.width.data == 3
    assert bv_type.value_range() == (0, 2**3)

    with pytest.raises(
        VerifyException,
        match="BitVectorType width must be strictly greater than zero, got 0",
    ):
        BitVectorType(0)


def test_function_type():
    func_type = FuncType([BoolType(), BitVectorType(32)], BoolType())
    assert list(func_type.domain_types) == [BoolType(), BitVectorType(32)]
    assert func_type.range_type == BoolType()


def test_not_op():
    arg = create_ssa_value(BoolType())
    op = NotOp(arg)
    assert op.result.type == BoolType()
    assert op.input == arg


@pytest.mark.parametrize("op_type", [AndOp, OrOp, XOrOp])
def test_variadic_bool_op(op_type: type[VariadicBoolOp]):
    arg1 = create_ssa_value(BoolType())
    arg2 = create_ssa_value(BoolType())
    arg3 = create_ssa_value(BoolType())
    op = op_type(arg1, arg2, arg3)
    assert op.result.type == BoolType()
    assert list(op.inputs) == [arg1, arg2, arg3]


@pytest.mark.parametrize("op_type", [EqOp, DistinctOp])
@pytest.mark.parametrize("type", [BoolType(), BitVectorType(32)])
def test_eq_distinct_op(op_type: type[VariadicBoolOp], type: BoolType | BitVectorType):
    arg1 = create_ssa_value(type)
    arg2 = create_ssa_value(type)
    arg3 = create_ssa_value(type)
    op = op_type(arg1, arg2, arg3)
    assert op.result.type == BoolType()
    assert list(op.inputs) == [arg1, arg2, arg3]


def test_implies_op():
    arg1 = create_ssa_value(BoolType())
    arg2 = create_ssa_value(BoolType())
    op = ImpliesOp(arg1, arg2)
    assert op.result.type == BoolType()
    assert op.lhs == arg1
    assert op.rhs == arg2


@pytest.mark.parametrize("op_type", [ExistsOp, ForallOp])
def test_quantifier_op(op_type: type[QuantifierOp]):
    arg1 = create_ssa_value(BoolType())
    region = Region([Block()])
    region.block.add_op(YieldOp(arg1))
    op = op_type(body=region)
    assert op.returned_value == arg1


def test_declare_fun():
    op = DeclareFunOp(BoolType(), "foo")
    assert op.name_prefix == StringAttr("foo")
    assert op.result.type == BoolType()

    op = DeclareFunOp(BoolType())
    assert op.name_prefix is None
    assert op.result.type == BoolType()

    op = DeclareFunOp(BitVectorType(32))
    assert op.name_prefix is None
    assert op.result.type == BitVectorType(32)


def test_apply_func():
    func = create_ssa_value(FuncType([BoolType(), BitVectorType(32)], BoolType()))
    arg1 = create_ssa_value(BoolType())
    arg2 = create_ssa_value(BitVectorType(32))
    op = ApplyFuncOp(func, arg1, arg2)

    assert op.result.type == BoolType()


def test_ite():
    arg1 = create_ssa_value(BoolType())
    arg2 = create_ssa_value(BitVectorType(32))
    arg3 = create_ssa_value(BitVectorType(32))
    op = IteOp(arg1, arg2, arg3)
    assert op.result.type == BitVectorType(32)
    assert op.cond == arg1
    assert op.then_value == arg2
    assert op.else_value == arg3


def test_assert_op():
    arg1 = create_ssa_value(BoolType())
    assert_op = AssertOp(arg1)
    assert assert_op.input == arg1


def test_bv_attr():
    bv_attr = BitVectorAttr(0, 32)
    assert bv_attr.value.data == 0
    assert bv_attr.type == BitVectorType(32)

    with pytest.raises(VerifyException, match="is out of range"):
        bv_attr = BitVectorAttr(-1, 32)

    with pytest.raises(VerifyException, match="is out of range"):
        bv_attr = BitVectorAttr(2**32, 32)


def test_bv_constant_op():
    bv_attr = BitVectorAttr(42, 32)
    op = BvConstantOp(bv_attr)
    assert op.value == bv_attr
    assert op.result.type == BitVectorType(32)

    op2 = BvConstantOp.from_value_and_type(42, 32)
    assert op2.value == bv_attr
    assert op2.result.type == BitVectorType(32)

    op3 = BvConstantOp.from_value_and_type(42, BitVectorType(32))
    assert op3.value == bv_attr
    assert op3.result.type == BitVectorType(32)


@pytest.mark.parametrize("op_type", [BVNotOp, BVNegOp])
def test_bv_unary_op(op_type: type[UnaryBVOp]):
    arg = create_ssa_value(BitVectorType(32))
    op = op_type(arg)
    assert op.input == arg
    assert op.result.type == arg.type


@pytest.mark.parametrize(
    "op_type",
    [
        BVAndOp,
        BVOrOp,
        BVXOrOp,
        BVAddOp,
        BVMulOp,
        BVUDivOp,
        BVSDivOp,
        BVURemOp,
        BVSRemOp,
        BVSModOp,
        BVShlOp,
        BVLShrOp,
        BVAShrOp,
    ],
)
def test_bv_binary_op(op_type: type[BinaryBVOp]):
    arg1 = create_ssa_value(BitVectorType(32))
    arg2 = create_ssa_value(BitVectorType(32))
    op = op_type(arg1, arg2)
    assert op.lhs == arg1
    assert op.rhs == arg2
    assert op.result.type == BitVectorType(32)



import marimo

__generated_with = "0.13.6"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo
    from xdsl.ir import Dialect
    from collections import defaultdict
    from xdsl.dialects import builtin, func, arith, scf
    from xdsl.context import Context
    from xdsl.parser import Parser
    from xdsl.utils import marimo as xmo
    from collections import Counter
    from xdsl.ir import OpResult, BlockArgument
    return (
        Context,
        Counter,
        Dialect,
        OpResult,
        Parser,
        arith,
        builtin,
        defaultdict,
        func,
        mo,
        scf,
        xmo,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# Traversing IR""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Parsing IR""")
    return


@app.cell(hide_code=True)
def _():
    sum_of_squares_text = """\
    func.func @sum_of_squares(%n: index) -> index {
      %zero = arith.constant 0 : index
      %one = arith.constant 1 : index
      %res = scf.for %i = %zero to %n step %one iter_args(%acc_in = %zero) -> (index) {
        %square = arith.muli %i, %i : index
        %acc_out = arith.addi %acc_in, %square : index
        scf.yield %acc_out : index
      }
      func.return %res : index
    }\
    """
    return (sum_of_squares_text,)


@app.cell(hide_code=True)
def _(mo, sum_of_squares_text):
    mo.md(
        fr"""
    In this notebook, we'll be looking at the structure of the following module:

    {mo.ui.code_editor(sum_of_squares_text, language="javascript", disabled=True)}
    """
    )
    return


@app.cell
def _(Context, arith, builtin, func, scf):
    # The context stores the available abstractions
    ctx = Context()
    ctx.load_dialect(builtin.Builtin)
    ctx.load_dialect(func.Func)
    ctx.load_dialect(arith.Arith)
    ctx.load_dialect(scf.Scf)
    return (ctx,)


@app.cell
def _(Parser, ctx, sum_of_squares_text, xmo):
    sum_of_squares_module = Parser(ctx, sum_of_squares_text).parse_module()

    # We can then parse and reprint the same module
    xmo.module_html(sum_of_squares_module)
    return (sum_of_squares_module,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Walking the IR""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    The Operation class exposes a `walk` function, which lets us iterate over the IR from top to bottom.

    For example, to get the names of all the operations in our module, we can write a snippet like this:
    """
    )
    return


@app.cell
def _(sum_of_squares_module):
    print([op.name for op in sum_of_squares_module.walk()])
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    These helpers will be useful for the exercises below:

    `Operation`:

    * `results: Sequence[OpResult]`
    * `operands: Sequence[SSAValue]`

    `Block`:

    * `parent_op: Operation | None`
    * `args: Sequence[BlockArgument]`

    `SSAValue`:

    * `uses: set[Use]`
    * `owner: Operation | Block`

    `OpResult` and `BlockArgument` are subclasses of `SSAValue`.

    `Use`:

    * `operation: Operation`
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Exercises""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""The rest of the notebook is a series of exercises for you to get acquainted with the API of xDSL.""")
    return


@app.cell(hide_code=True)
def _(mo, sum_of_squares_module, unique_operations):
    try:
        _sorted = "{" + ", ".join(sorted(unique_operations(sum_of_squares_module))) + "}"
        _res = str(_sorted)
    except Exception as e:
        _res = f"{type(e).__name__}: {e}"

    mo.md(
        fr"""
        ### Exercise 1. Unique Operations

        Modify the function below to return the set of operations used in the module.

        ```
        Expected: {{arith.addi, arith.constant, arith.muli, builtin.module, func.func, func.return, scf.for, scf.yield}}
        Result:   {_res}
        ```
        """
    )
    return


@app.cell
def _(builtin):
    def unique_operations(module: builtin.ModuleOp) -> set[str]:
        return set()
    return (unique_operations,)


@app.cell(hide_code=True)
def _(builtin):
    # Solution

    def _unique_operations(module: builtin.ModuleOp) -> set[str]:
        return {op.name for op in module.walk()}
    return


@app.cell(hide_code=True)
def _(mo, operation_counts, sum_of_squares_module):
    try:
        _res = str(operation_counts(sum_of_squares_module))
    except Exception as e:
        _res = f"{type(e).__name__}: {e}"

    mo.md(
        fr"""
        ### Exercise 2. Operation Counter

        Modify the function below to return the number of instances of the operations in the module.

        ```
        Expected: {{'builtin.module': 1, 'func.func': 1, 'arith.constant': 2, 'scf.for': 1, 'arith.muli': 1, 'arith.addi': 1, 'scf.yield': 1, 'func.return': 1}}
        Result:   {_res}
        ```
        """
    )
    return


@app.cell
def _(builtin):
    # `Counter` might come in handy

    def operation_counts(module: builtin.ModuleOp) -> dict[str, int]:
        return {}
    return (operation_counts,)


@app.cell(hide_code=True)
def _(Counter, builtin):
    # solution

    def _operation_counts(module: builtin.ModuleOp) -> dict[str, int]:
        return dict(Counter(op.name for op in module.walk()))
    return


@app.cell(hide_code=True)
def _(mo, operations_by_dialect, sum_of_squares_module):
    try:
        _unsorted = operations_by_dialect(sum_of_squares_module)
        _sorted = {
            k: sorted(_unsorted[k])
            for k in sorted(_unsorted)
        }
        _res = str(_sorted)
    except Exception as e:
        _res = f"{type(e).__name__}: {e}"

    mo.md(
        fr"""
        ### Exercise 3. Operations By Dialect

        Modify the function below to return the operations by dialect in the module

        ```
        Expected: {{'arith': ['addi', 'constant', 'constant', 'muli'], 'builtin': ['module'], 'func': ['func', 'return'], 'scf': ['for', 'yield']}}
        Result:   {_res}
        ```
        """
    )
    return


@app.cell
def _(builtin):
    # `defaultdict` might come in handy

    def operations_by_dialect(module: builtin.ModuleOp) -> dict[str, list[str]]:
        return {}
    return (operations_by_dialect,)


@app.cell(hide_code=True)
def _(Dialect, builtin, defaultdict):
    # solution

    def _operations_by_dialect(module: builtin.ModuleOp) -> dict[str, list[str]]:
        res = defaultdict(list)

        for op in module.walk():
            d, o = Dialect.split_name(op.name)
            res[d].append(o)
        return dict(res)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## SSA Values""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    SSA values in IR are either defined as results or block arguments.
    In xDSL, the values have optional name hints, and are numbered in ascending order when printing if the name hint is missing.
    """
    )
    return


@app.cell(hide_code=True)
def _(sum_of_squares_module):
    all_operands = set(val.name_hint for op in sum_of_squares_module.walk() for val in op.operands)
    all_results = set(val.name_hint for op in sum_of_squares_module.walk() for val in op.results)
    all_block_args = set(
        val.name_hint
        for op in sum_of_squares_module.walk()
        for region in op.regions
        for block in region.blocks
        for val in block.args
    )
    all_ssa_values = all_operands | all_results | all_block_args
    return (all_ssa_values,)


@app.cell(hide_code=True)
def _(all_ssa_values, mo):
    mo.md(
        fr"""
    Here are all the name hints of SSA values in our module:

    ```
    {{{', '.join(sorted(all_ssa_values))}}}
    ```
    """
    )
    return


@app.cell(hide_code=True)
def _(definition_by_use, mo, sum_of_squares_module):
    try:
        _unsorted = definition_by_use(sum_of_squares_module)
        _sorted = {
            k: sorted(_unsorted[k])
            for k in sorted(_unsorted)
        }
        _res = str(_sorted)
    except Exception as e:
        _res = f"{type(e).__name__}: {e}"

    mo.md(
        fr"""
        ### Exercise 4. Definition By Use

        Modify the function below to return the operation that defines the value by the operation that uses it.
        If the operand is a block argument, use the name of the parent operation.

        Note: the `SSAValue` class provides a `owner` helper that might be useful.


        ```
        Expected: {{'arith.addi': ['arith.muli', 'scf.for'], 'arith.muli': ['scf.for', 'scf.for'], 'func.return': ['scf.for'], 'scf.for': ['arith.constant', 'arith.constant', 'arith.constant', 'func.func'], 'scf.yield': ['arith.addi']}}
        Result:   {_res}
        ```
        """
    )
    return


@app.cell
def _(builtin):
    def definition_by_use(module: builtin.ModuleOp) -> dict[str, list[str]]:
        return {}
    return (definition_by_use,)


@app.cell(hide_code=True)
def _(OpResult, builtin, defaultdict):
    # Solution

    def _definition_by_use(module: builtin.ModuleOp) -> dict[str, list[str]]:
        res = defaultdict(list)

        for use_op in module.walk():
            for operand in use_op.operands:
                if isinstance(operand, OpResult):
                    op = operand.op
                else:
                    op = operand.block.parent_op()

                res[use_op.name].append(op.name)
        return dict(res)
    return


@app.cell(hide_code=True)
def _(mo, sum_of_squares_module, uses_by_definition):
    try:
        _unsorted = uses_by_definition(sum_of_squares_module)
        _sorted = {
            k: sorted(_unsorted[k])
            for k in sorted(_unsorted)
        }
        _res = str(_sorted)
    except Exception as e:
        _res = f"{type(e).__name__}: {e}"

    mo.md(
        fr"""
        ### Exercise 5. Uses By Definition

        Modify the function below to return the operations that use the result by the operation that defines it.

        Note: the `SSAValue` class provides a `uses` helper that might be useful.

        ```
        Expected: {{'arith.addi': ['scf.yield'], 'arith.constant': ['scf.for', 'scf.for', 'scf.for'], 'arith.muli': ['arith.addi'], 'scf.for': ['func.return']}}
        Result:   {_sorted}
        ```
        """
    )
    return


@app.cell
def _(builtin):
    def uses_by_definition(module: builtin.ModuleOp) -> dict[str, list[str]]:
        return {}
    return (uses_by_definition,)


@app.cell(hide_code=True)
def _(builtin, defaultdict):
    # Solution

    def _uses_by_definition(module: builtin.ModuleOp) -> dict[str, list[str]]:
        res = defaultdict(list)

        for def_op in module.walk():
            for result in def_op.results:
                for use in result.uses:
                    res[def_op.name].append(use.operation.name)

        return dict(res)
    return


if __name__ == "__main__":
    app.run()



from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import builtin, func, transform
from xdsl.interpreter import Interpreter
from xdsl.interpreters.transform import TransformFunctions
from xdsl.ir import Block, Region
from xdsl.parser import Parser
from xdsl.transforms import get_all_passes


def test_empty_transform_module():
    payload = """
    module {
        func.func @foo() {
            func.return
        }
    }
    """

    ty = transform.OperationType("builtin.module")
    block = Block(arg_types=[ty])
    with ImplicitBuilder(block):
        transform.YieldOp(block.args[0])

    module = builtin.ModuleOp(
        [], attributes={"transform.with_named_sequence": builtin.UnitAttr()}
    )
    with ImplicitBuilder(module.body):
        body = Region(block)
        sym_name = "__transform_main"
        function_type = builtin.FunctionType.from_lists([ty], [ty])
        named_sequence = transform.NamedSequenceOp(sym_name, function_type, body)

    ctx = Context()
    ctx.load_dialect(builtin.Builtin)
    ctx.load_dialect(func.Func)
    ctx.load_dialect(transform.Transform)

    interpreter = Interpreter(module)
    interpreter.register_implementations(TransformFunctions(ctx, get_all_passes()))

    expected = Parser(ctx, payload).parse_module()
    (observed,) = interpreter.call_op(named_sequence, (expected,))
    assert expected is observed



from xdsl.context import Context
from xdsl.interpreter import Interpreter
from xdsl.interpreters import (
    affine,
    arith,
    builtin,
    cf,
    func,
    linalg,
    memref,
    memref_stream,
    ml_program,
    pdl,
    printf,
    riscv,
    riscv_cf,
    riscv_debug,
    riscv_func,
    riscv_libc,
    riscv_scf,
    riscv_snitch,
    scf,
    snitch_stream,
    tensor,
)


def register_implementations(interpreter: Interpreter, ctx: Context):
    interpreter.register_implementations(affine.AffineFunctions())
    interpreter.register_implementations(arith.ArithFunctions())
    interpreter.register_implementations(builtin.BuiltinFunctions())
    interpreter.register_implementations(cf.CfFunctions())
    interpreter.register_implementations(func.FuncFunctions())
    interpreter.register_implementations(linalg.LinalgFunctions())
    interpreter.register_implementations(memref_stream.MemRefStreamFunctions())
    interpreter.register_implementations(memref.MemRefFunctions())
    interpreter.register_implementations(ml_program.MLProgramFunctions())
    interpreter.register_implementations(pdl.PDLRewriteFunctions(ctx))
    interpreter.register_implementations(printf.PrintfFunctions())
    interpreter.register_implementations(riscv_cf.RiscvCfFunctions())
    interpreter.register_implementations(riscv_debug.RiscvDebugFunctions())
    interpreter.register_implementations(riscv_func.RiscvFuncFunctions())
    interpreter.register_implementations(riscv_libc.RiscvLibcFunctions())
    interpreter.register_implementations(riscv_scf.RiscvScfFunctions())
    interpreter.register_implementations(riscv_snitch.RiscvSnitchFunctions())
    interpreter.register_implementations(riscv.RiscvFunctions())
    interpreter.register_implementations(scf.ScfFunctions())
    interpreter.register_implementations(snitch_stream.SnitchStreamFunctions())
    interpreter.register_implementations(tensor.TensorFunctions())



from __future__ import annotations

from abc import ABC
from collections.abc import Sequence
from typing import ClassVar, cast

from xdsl.dialects.builtin import (
    I1,
    AffineMapAttr,
    AnyFloat,
    AnyFloatConstr,
    ArrayAttr,
    BoolAttr,
    DenseArrayBase,
    IndexType,
    IndexTypeConstr,
    IntegerType,
    MemRefType,
    SignlessIntegerConstraint,
    TensorType,
    VectorBaseTypeAndRankConstraint,
    VectorBaseTypeConstraint,
    VectorRankConstraint,
    VectorType,
    i1,
    i64,
)
from xdsl.dialects.utils import (
    get_dynamic_index_list,
    split_dynamic_index_list,
    verify_dynamic_index_list,
)
from xdsl.dialects.utils.dynamic_index_list import DynamicIndexList
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.ir.affine import AffineConstantExpr, AffineDimExpr, AffineMap
from xdsl.irdl import (
    AttrSizedOperandSegments,
    IRDLOperation,
    ParsePropInAttrDict,
    VarConstraint,
    base,
    irdl_op_definition,
    operand_def,
    opt_operand_def,
    opt_prop_def,
    opt_result_def,
    prop_def,
    result_def,
    traits_def,
    var_operand_def,
)
from xdsl.parser import Parser, UnresolvedOperand
from xdsl.printer import Printer
from xdsl.traits import Pure
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa
from xdsl.utils.lexer import Position

DYNAMIC_INDEX: int = -(2**63)


@irdl_op_definition
class LoadOp(IRDLOperation):
    name = "vector.load"
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)
    result = result_def(VectorType)
    nontemporal = opt_prop_def(BoolAttr, default_value=BoolAttr.from_bool(False))

    irdl_options = [ParsePropInAttrDict()]
    assembly_format = (
        "$base `[` $indices `]` attr-dict `:` type($base) `,` type($result)"
    )

    def verify_(self):
        assert isa(self.base.type, MemRefType)
        assert isa(self.result.type, VectorType[Attribute])

        if self.base.type.element_type != self.result.type.element_type:
            raise VerifyException(
                "MemRef element type should match the Vector element type."
            )

        if self.base.type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each dimension.")

    @staticmethod
    def get(
        ref: SSAValue | Operation, indices: Sequence[SSAValue | Operation]
    ) -> LoadOp:
        ref = SSAValue.get(ref, type=MemRefType)

        return LoadOp.build(
            operands=[ref, indices],
            result_types=[VectorType(ref.type.element_type, [1])],
        )


@irdl_op_definition
class StoreOp(IRDLOperation):
    name = "vector.store"
    vector = operand_def(VectorType)
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)
    nontemporal = opt_prop_def(BoolAttr, default_value=BoolAttr.from_bool(False))

    irdl_options = [ParsePropInAttrDict()]
    assembly_format = (
        "$vector `,` $base `[` $indices `]` attr-dict `:` type($base) `,` type($vector)"
    )

    def verify_(self):
        assert isa(self.base.type, MemRefType)
        assert isa(self.vector.type, VectorType[Attribute])

        if self.base.type.element_type != self.vector.type.element_type:
            raise VerifyException(
                "MemRef element type should match the Vector element type."
            )

        if self.base.type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each dimension.")

    @staticmethod
    def get(
        vector: Operation | SSAValue,
        ref: Operation | SSAValue,
        indices: Sequence[Operation | SSAValue],
    ) -> StoreOp:
        return StoreOp.build(operands=[vector, ref, indices])


@irdl_op_definition
class BroadcastOp(IRDLOperation):
    """
    See external [documentation](https://mlir.llvm.org/docs/Dialects/Vector/#vectorbroadcast-vectorbroadcastop).
    """

    name = "vector.broadcast"
    source = operand_def()
    vector = result_def(VectorType)
    traits = traits_def(Pure())

    assembly_format = "$source attr-dict `:` type($source) `to` type($vector)"

    def verify_(self):
        if isa(self.source.type, VectorType):
            element_type = self.source.type.element_type
        else:
            element_type = self.source.type

        if element_type != self.vector.type.element_type:
            raise VerifyException(
                "Source operand and result vector must have the same element type."
            )

    @staticmethod
    def get(source: Operation | SSAValue) -> BroadcastOp:
        return BroadcastOp.build(
            operands=[source],
            result_types=[VectorType(SSAValue.get(source).type, [1])],
        )


@irdl_op_definition
class FMAOp(IRDLOperation):
    name = "vector.fma"

    T: ClassVar = VarConstraint("T", VectorType.constr(AnyFloatConstr))

    lhs = operand_def(T)
    rhs = operand_def(T)
    acc = operand_def(T)
    res = result_def(T)
    traits = traits_def(Pure())

    assembly_format = "$lhs `,` $rhs `,` $acc attr-dict `:` type($lhs)"

    @staticmethod
    def get(
        lhs: Operation | SSAValue, rhs: Operation | SSAValue, acc: Operation | SSAValue
    ) -> FMAOp:
        lhs = SSAValue.get(lhs, type=VectorType)

        return FMAOp.build(
            operands=[lhs, rhs, acc],
            result_types=[VectorType(lhs.type.element_type, [1])],
        )


@irdl_op_definition
class MaskedLoadOp(IRDLOperation):
    name = "vector.maskedload"
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)
    mask = operand_def(VectorBaseTypeAndRankConstraint(i1, 1))
    pass_thru = operand_def(VectorType)
    result = result_def(VectorRankConstraint(1))

    assembly_format = "$base `[` $indices `]` `,` $mask `,` $pass_thru attr-dict `:` type($base) `,` type($mask) `,` type($pass_thru) `into` type($result)"  # noqa: E501

    def verify_(self):
        memref_type = self.base.type
        assert isa(memref_type, MemRefType)
        memref_element_type = memref_type.element_type

        res_type = self.result.type
        assert isa(res_type, VectorType[Attribute])
        res_element_type = res_type.element_type

        passthrough_type = self.pass_thru.type
        assert isa(passthrough_type, VectorType[Attribute])
        passthrough_element_type = passthrough_type.element_type

        if memref_element_type != res_element_type:
            raise VerifyException(
                "MemRef element type should match the result vector and passthrough vector "
                "element type. Found different element types for memref and result."
            )
        elif memref_element_type != passthrough_element_type:
            raise VerifyException(
                "MemRef element type should match the result vector and passthrough vector "
                "element type. Found different element types for memref and passthrough."
            )

        if memref_type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each memref dimension.")

    @staticmethod
    def get(
        memref: SSAValue | Operation,
        indices: Sequence[SSAValue | Operation],
        mask: SSAValue | Operation,
        passthrough: SSAValue | Operation,
    ) -> MaskedLoadOp:
        memref = SSAValue.get(memref, type=MemRefType)

        return MaskedLoadOp.build(
            operands=[memref, indices, mask, passthrough],
            result_types=[VectorType(memref.type.element_type, [1])],
        )


@irdl_op_definition
class MaskedStoreOp(IRDLOperation):
    name = "vector.maskedstore"
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)
    mask = operand_def(VectorBaseTypeAndRankConstraint(i1, 1))
    value_to_store = operand_def(VectorRankConstraint(1))

    assembly_format = "$base `[` $indices `]` `,` $mask `,` $value_to_store attr-dict `:` type($base) `,` type($mask) `,` type($value_to_store)"  # noqa: E501

    def verify_(self):
        memref_type = self.base.type
        assert isa(memref_type, MemRefType)
        memref_element_type = memref_type.element_type

        value_to_store_type = self.value_to_store.type
        assert isa(value_to_store_type, VectorType[Attribute])

        mask_type = self.mask.type
        assert isa(mask_type, VectorType[Attribute])

        if memref_element_type != value_to_store_type.element_type:
            raise VerifyException(
                "MemRef element type should match the stored vector type. "
                "Obtained types were "
                + str(memref_element_type)
                + " and "
                + str(value_to_store_type.element_type)
                + "."
            )

        if memref_type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each memref dimension.")

    @staticmethod
    def get(
        memref: SSAValue | Operation,
        indices: Sequence[SSAValue | Operation],
        mask: SSAValue | Operation,
        value_to_store: SSAValue | Operation,
    ) -> MaskedStoreOp:
        return MaskedStoreOp.build(operands=[memref, indices, mask, value_to_store])


@irdl_op_definition
class PrintOp(IRDLOperation):
    name = "vector.print"
    source = operand_def()

    @staticmethod
    def get(source: Operation | SSAValue) -> PrintOp:
        return PrintOp.build(operands=[source])


@irdl_op_definition
class CreateMaskOp(IRDLOperation):
    name = "vector.create_mask"
    mask_dim_sizes = var_operand_def(IndexType)
    mask_vector = result_def(VectorBaseTypeConstraint(i1))

    assembly_format = "$mask_dim_sizes attr-dict `:` type(results)"

    def verify_(self):
        assert isa(self.mask_vector.type, VectorType[Attribute])
        if self.mask_vector.type.get_num_dims() != len(self.mask_dim_sizes):
            raise VerifyException(
                "Expected an operand value for each dimension of resultant mask."
            )

    @staticmethod
    def get(mask_operands: list[Operation | SSAValue]) -> CreateMaskOp:
        return CreateMaskOp.build(
            operands=[mask_operands],
            result_types=[VectorType(i1, [1])],
        )


@irdl_op_definition
class ExtractOp(IRDLOperation):
    name = "vector.extract"

    _T: ClassVar = VarConstraint(
        "T", base(IntegerType) | base(IndexType) | AnyFloatConstr
    )
    _V: ClassVar = VarConstraint("V", VectorType.constr(_T))

    static_position = prop_def(DenseArrayBase.constr(i64))

    vector = operand_def(_V)
    dynamic_position = var_operand_def(IndexTypeConstr)

    result = result_def(VectorType.constr(_T) | _T)

    traits = traits_def(Pure())

    DYNAMIC_INDEX: ClassVar = DYNAMIC_INDEX
    """This value is used to indicate that a position is a dynamic index."""

    assembly_format = (
        "$vector `` custom<DynamicIndexList>($dynamic_position, $static_position)"
        " attr-dict `:` type($result) `from` type($vector)"
    )

    custom_directives = (DynamicIndexList,)

    def get_mixed_position(self) -> list[SSAValue | int]:
        """
        Returns the list of positions, represented as either an SSAValue or an int
        """
        static_positions = self.static_position.get_values()
        return get_dynamic_index_list(
            static_positions,
            self.dynamic_position,
            ExtractOp.DYNAMIC_INDEX,
        )

    def verify_(self):
        # Check that static position attribute and dynamic position operands
        # are compatible.
        static_values = self.static_position.get_values()
        verify_dynamic_index_list(
            static_values,
            self.dynamic_position,
            self.DYNAMIC_INDEX,
        )

        num_indices = len(self.static_position)
        vector_type = self.vector.type
        assert isa(vector_type, VectorType[Attribute])
        # Check that the number of dimensions match
        if isa(self.result.type, VectorType):
            if (
                num_indices + self.result.type.get_num_dims()
                != vector_type.get_num_dims()
            ):
                raise VerifyException(
                    f"Expected position attribute rank ({num_indices}) + result rank "
                    f"({self.result.type.get_num_dims()}) to "
                    f"match source vector rank ({vector_type.get_num_dims()})."
                )
        else:
            if num_indices != vector_type.get_num_dims():
                raise VerifyException(
                    f"Expected position attribute rank ({num_indices}) to match "
                    f"source vector rank ({vector_type.get_num_dims()})."
                )

    def __init__(
        self,
        vector: SSAValue,
        positions: Sequence[SSAValue | int],
        result_type: Attribute,
    ):
        static_positions, dynamic_positions = split_dynamic_index_list(
            positions, ExtractOp.DYNAMIC_INDEX
        )

        super().__init__(
            operands=[vector, dynamic_positions],
            result_types=[result_type],
            properties={
                "static_position": DenseArrayBase.from_list(i64, static_positions)
            },
        )


@irdl_op_definition
class ExtractElementOp(IRDLOperation):
    name = "vector.extractelement"
    vector = operand_def(VectorType)
    position = opt_operand_def(IndexTypeConstr | SignlessIntegerConstraint)
    result = result_def(Attribute)
    traits = traits_def(Pure())

    def verify_(self):
        assert isa(self.vector.type, VectorType[Attribute])

        if self.result.type != self.vector.type.element_type:
            raise VerifyException(
                "Expected result type to match element type of vector operand."
            )

        if self.vector.type.get_num_dims() == 0:
            if self.position is not None:
                raise VerifyException("Expected position to be empty with 0-D vector.")
            return
        if self.vector.type.get_num_dims() != 1:
            raise VerifyException("Unexpected >1 vector rank.")
        if self.position is None:
            raise VerifyException("Expected position for 1-D vector.")

    def __init__(
        self,
        vector: SSAValue | Operation,
        position: SSAValue | Operation | None = None,
    ):
        vector = SSAValue.get(vector, type=VectorType)

        result_type = vector.type.element_type

        super().__init__(
            operands=[vector, position],
            result_types=[result_type],
        )


@irdl_op_definition
class InsertOp(IRDLOperation):
    name = "vector.insert"

    _T: ClassVar = VarConstraint(
        "T", base(IntegerType) | base(IndexType) | AnyFloatConstr
    )
    _V: ClassVar = VarConstraint("V", VectorType.constr(_T))

    static_position = prop_def(DenseArrayBase.constr(i64))

    source = operand_def(VectorType.constr(_T) | _T)
    dest = operand_def(_V)
    dynamic_position = var_operand_def(IndexTypeConstr)

    result = result_def(_V)

    traits = traits_def(Pure())

    DYNAMIC_INDEX: ClassVar = -(2**63)
    """This value is used to indicate that a position is a dynamic index."""

    assembly_format = (
        "$source `,` $dest custom<DynamicIndexList>($dynamic_position, $static_position)"
        "attr-dict `:` type($source) `into` type($dest)"
    )

    custom_directives = (DynamicIndexList,)

    def get_mixed_position(self) -> list[SSAValue | int]:
        """
        Returns the list of positions, represented as either an SSAValue or an int.
        """
        static_positions = self.static_position.get_values()
        return get_dynamic_index_list(
            static_positions,
            self.dynamic_position,
            InsertOp.DYNAMIC_INDEX,
        )

    def verify_(self):
        # Check that static position attribute and dynamic position operands
        # are compatible.
        static_values = self.static_position.get_values()
        verify_dynamic_index_list(
            static_values,
            self.dynamic_position,
            self.DYNAMIC_INDEX,
        )

        num_indices = len(self.static_position)
        # Check that the number of dimensions match
        if isa(self.source.type, VectorType):
            if (
                num_indices + self.source.type.get_num_dims()
                != self.result.type.get_num_dims()
            ):
                raise VerifyException(
                    f"Expected position attribute rank ({num_indices}) + source rank "
                    f"({self.source.type.get_num_dims()}) to "
                    f"match dest vector rank ({self.result.type.get_num_dims()})."
                )
        else:
            if num_indices != self.result.type.get_num_dims():
                raise VerifyException(
                    f"Expected position attribute rank ({num_indices}) to match "
                    f"dest vector rank ({self.result.type.get_num_dims()})."
                )

    def __init__(
        self,
        source: SSAValue,
        dest: SSAValue,
        positions: Sequence[SSAValue | int],
        result_type: Attribute | None = None,
    ):
        static_positions, dynamic_positions = split_dynamic_index_list(
            positions, InsertOp.DYNAMIC_INDEX
        )

        if result_type is None:
            result_type = dest.type

        super().__init__(
            operands=[source, dest, dynamic_positions],
            result_types=[result_type],
            properties={
                "static_position": DenseArrayBase.from_list(i64, static_positions)
            },
        )


@irdl_op_definition
class InsertElementOp(IRDLOperation):
    name = "vector.insertelement"
    source = operand_def(Attribute)
    dest = operand_def(VectorType)
    position = opt_operand_def(IndexTypeConstr | SignlessIntegerConstraint)
    result = result_def(VectorType)
    traits = traits_def(Pure())

    def verify_(self):
        assert isa(self.dest.type, VectorType[Attribute])

        if self.result.type != self.dest.type:
            raise VerifyException(
                "Expected dest operand and result to have matching types."
            )
        if self.source.type != self.dest.type.element_type:
            raise VerifyException(
                "Expected source operand type to match element type of dest operand."
            )

        if self.dest.type.get_num_dims() == 0:
            if self.position is not None:
                raise VerifyException("Expected position to be empty with 0-D vector.")
            return
        if self.dest.type.get_num_dims() != 1:
            raise VerifyException("Unexpected >1 vector rank.")
        if self.position is None:
            raise VerifyException("Expected position for 1-D vector.")

    def __init__(
        self,
        source: SSAValue | Operation,
        dest: SSAValue | Operation,
        position: SSAValue | Operation | None = None,
    ):
        dest = SSAValue.get(dest, type=VectorType)

        result_type = SSAValue.get(dest).type

        super().__init__(
            operands=[source, dest, position],
            result_types=[result_type],
        )


class VectorTransferOperation(IRDLOperation, ABC):
    """
    Encodes properties of a `vector.transfer_read` or `vector.transfer_write`
    operation. Vector transfer ops have:

    - A shaped value that the op reads from/writes to: a memref or a tensor.
    - A vector, either as a result or as an operand.
    - Indices that describe where the transfer from/to the shaped value starts.
    - An optional mask.
    - An optional in_bounds array to indicate transfer dimensions that are
      guaranteed to be in-bounds.
    - A permutation map to indicate transposes and broadcasts.

    The "vector rank" is the rank of the vector type. E.g.:
    ```mlir
    // Transfer with shaped value rank 2 and vector (transfer) rank 1.
    %0 = vector.transfer_read %arg0[%c3, %c3], %f0
        {permutation_map = affine_map<(d0, d1) -> (d0)>}
        : memref<?x?xf32>, vector<128xf32>
    ```

    The "vector transfer rank" is the number of dimensions that participate in
    the transfer and broadcasts, and matches the number of results in the
    permutation map. In most cases, the vector rank matches the vector transfer
    rank; the only exception is when a vector is flattened as part of the
    transfer (see `permutation_map`).

    Mirrors VectorTransferOpInterface from [MLIR](https://github.com/llvm/llvm-project/blob/main/mlir/include/mlir/Interfaces/VectorInterfaces.td)
    """

    permutation_map = prop_def(AffineMapAttr)
    """
    The permutation map that describes the mapping of vector
    dimensions to source dimensions, as well as broadcast dimensions.

    The permutation result has one result per vector transfer dimension.
    Each result is either a dim expression, indicating the corresponding
    dimension in the source operand, or a constant "0" expression,
    indicating a broadcast dimension.

    Note: Nested vector dimensions that are flattened by this op are not
    accounted for in the permutation map. E.g.:
    ```mlir
    // Vector type has rank 4, but permutation map has only 2 results. That
    // is because there are only 2 transfer dimensions.
    %0 = vector.transfer_read %arg1[%c3, %c3], %vf0
        {permutation_map = affine_map<(d0, d1) -> (d0, d1)>}
        : memref<?x?xvector<4x3xf32>>, vector<1x1x4x3xf32>
    ```
    """

    in_bounds = prop_def(ArrayAttr[BoolAttr])
    """
    For every vector dimension, the boolean array attribute `in_bounds` specifies if the
    transfer is guaranteed to be within the source bounds. If set to `“false”`, accesses
    (including the starting point) may run out-of-bounds along the respective vector
    dimension as the index increases. Non-vector dimensions must always be in-bounds.
    The `in_bounds` array length has to be equal to the vector rank. This attribute has
    a default value: `false` (i.e. “out-of-bounds”). When skipped in the textual IR, the
    default value is assumed. Similarly, the OP printer will omit this attribute when
    all dimensions are out-of-bounds (i.e. the default value is used).
    """

    @staticmethod
    def infer_transfer_op_mask_type(
        vec_type: VectorType, perm_map: AffineMap
    ) -> VectorType[I1]:
        """
        Given a resulting vector type and a permutation map from the dimensions of the
        shaped type to the vector type dimensions, return the vector type of the mask.
        """
        unused_dims_bit_vector = tuple(
            not dim for dim in perm_map.used_dims_bit_vector()
        )
        inv_perm_map = perm_map.drop_dims(unused_dims_bit_vector).inverse_permutation()
        assert inv_perm_map is not None, "Inversed permutation map couldn't be computed"
        mask_shape = inv_perm_map.eval(vec_type.get_shape(), ())
        scalable_dims = ArrayAttr(
            BoolAttr.from_bool(bool(b))
            for b in inv_perm_map.eval(vec_type.get_scalable_dims(), ())
        )
        res = VectorType(i1, mask_shape, scalable_dims)
        return res

    @staticmethod
    def get_transfer_minor_identity_map(
        shaped_type: TensorType | MemRefType, vector_type: VectorType
    ) -> AffineMap:
        """
        Get the minor identity map for a transfer operation.

        This is a helper function to compute the default permutation map for
        transfer operations when none is specified.
        """
        element_vector_rank = 0
        element_type = shaped_type.element_type
        if isa(element_type, VectorType):
            element_vector_rank += element_type.get_num_dims()

        # 0-d transfers are to/from tensor<t>/memref<t> and vector<1xt>.
        # TODO: replace once we have 0-d vectors.
        if shaped_type.get_num_dims() == 0 and vector_type.get_shape() == (1,):
            return AffineMap.constant_map(0)

        return AffineMap.minor_identity(
            shaped_type.get_num_dims(),
            vector_type.get_num_dims() - element_vector_rank,
        )

    def _print_attrs(self, printer: Printer):
        reserved_attr_names = {"operandSegmentSizes"}
        if self.permutation_map.data.is_minor_identity():
            reserved_attr_names.add("permutation_map")
        if not any(self.in_bounds):
            reserved_attr_names.add("in_bounds")
        printer.print_op_attributes(
            self.attributes | self.properties, reserved_attr_names=reserved_attr_names
        )

    @staticmethod
    def resolve_attrs(
        parser: Parser,
        attributes_dict: dict[str, Attribute],
        shaped_type: TensorType | MemRefType,
        vector_type: VectorType,
        mask_start_pos: Position | None,
        mask_end_pos: Position | None,
        mask: UnresolvedOperand | None,
        types_pos: Position,
    ):
        # Create default permutation_map if not provided in attributes
        permutation_map = None
        if attributes_dict and "permutation_map" in attributes_dict:
            permutation_map = attributes_dict["permutation_map"]
            assert isinstance(permutation_map, AffineMapAttr)
        else:
            # Create identity permutation map for the shaped type's rank
            permutation_map = AffineMapAttr(
                VectorTransferOperation.get_transfer_minor_identity_map(
                    shaped_type, vector_type
                )
            )

        # Create in_bounds attribute if not provided
        in_bounds = None
        if attributes_dict and "in_bounds" in attributes_dict:
            in_bounds = cast(ArrayAttr[BoolAttr], attributes_dict["in_bounds"])
        else:
            # Default: all dimensions are out-of-bounds
            in_bounds = ArrayAttr(
                (BoolAttr.from_bool(False),) * len(permutation_map.data.results)
            )

        if mask is not None:
            if isa(shaped_type.element_type, VectorType):
                assert mask_start_pos is not None
                assert mask_end_pos is not None
                parser.raise_error(
                    "does not support masks with vector element type",
                    at_position=mask_start_pos,
                    end_position=mask_end_pos,
                )
            if vector_type.get_num_dims() != len(permutation_map.data.results):
                parser.raise_error(
                    "expected the same rank for the vector and the "
                    "results of the permutation map",
                    types_pos,
                )
            # Instead of adding the mask type as an op type, compute it based on the
            # vector type and the permutation map (to keep the type signature small).
            mask_type = VectorTransferOperation.infer_transfer_op_mask_type(
                vector_type, permutation_map.data
            )
            resolved_mask = parser.resolve_operand(mask, mask_type)
        else:
            resolved_mask = None

        return resolved_mask, permutation_map, in_bounds

    def has_broadcast_dim(self):
        """
        Return "true" if at least one of the vector dimensions is a broadcasted dimension.
        """
        return any(
            isinstance(expr, AffineConstantExpr) and expr.value == 0
            for expr in self.permutation_map.data.results
        )

    @staticmethod
    def verify_op(
        op: TransferReadOp | TransferWriteOp,
        shaped_type: MemRefType | TensorType,
        vector_type: VectorType,
        mask_type: VectorType[I1] | None,
        inferred_mask_type: VectorType[I1] | None,
        permutation_map: AffineMap,
        in_bounds: ArrayAttr[BoolAttr],
    ):
        """
        This mirrors VectorOps.cpp -> verifyTransferOp from MLIR
        """

        element_type = shaped_type.element_type
        vector_element_type = vector_type.element_type

        if isa(element_type, VectorType):
            # Memref or tensor has vector element type
            # TODO verify vector element type
            pass
        else:
            # Memref of tensor has scalar element type
            if isa(vector_element_type, IndexType):
                if not isa(element_type, IndexType):
                    raise VerifyException(
                        "Element type of source is index, expected element type of vector also to be index"
                    )
            else:
                assert isa(vector_element_type, IntegerType | AnyFloat)
                assert isa(element_type, IntegerType | AnyFloat)

                minor_size = (
                    1
                    if vector_type.get_num_dims() == 0
                    else vector_type.get_shape()[-1]
                )
                result_vec_size = vector_element_type.bitwidth * minor_size
                if result_vec_size % element_type.bitwidth != 0:
                    raise VerifyException(
                        f'"{op.name}" requires the bitwidth of the minor 1-D vector to be '
                        "an integral multiple of the bitwidth of the source element type"
                    )

            # Check that permutation map results match rank of vector type.
            if len(permutation_map.results) != vector_type.get_num_dims():
                raise VerifyException(
                    f'"{op.name}" requires a permutation_map with result dims of the same rank as the vector type'
                )

        if permutation_map.num_symbols != 0:
            raise VerifyException(
                f'"{op.name}" requires permutation_map without symbols'
            )

        if permutation_map.num_dims != shaped_type.get_num_dims():
            raise VerifyException(
                f'"{op.name}" requires a permutation_map with input dims of the same rank as the source type'
            )

        if mask_type:
            if mask_type != inferred_mask_type:
                raise VerifyException(
                    f'"{op.name}" inferred mask type ({inferred_mask_type}) and mask operand type ({mask_type}) don\'t match'
                )

        if len(in_bounds) != len(permutation_map.results):
            raise VerifyException(
                f'"{op.name}" expects the in_bounds attr of same rank as permutation_map results: '
                f"{str(permutation_map)} vs in_bounds of of size {len(in_bounds)}"
            )

    @staticmethod
    def verify_permutation_map(
        op: TransferReadOp | TransferWriteOp,
        permutation_map: AffineMap,
    ):
        """
        This mirrors VectorOps.cpp -> verifyPermutationMap
        """

        seen: list[bool] = [False for _ in range(permutation_map.num_dims)]

        for expr in permutation_map.results:
            if isa(expr, AffineConstantExpr):
                if expr.value != 0:
                    raise VerifyException(
                        f'"{op.name}" requires a projected permutation_map '
                        "(at most one dim or the zero constant can appear in each result)"
                    )
                continue
            if not isa(expr, AffineDimExpr):
                raise VerifyException(
                    f'"{op.name}" requires a projected permutation_map '
                    "(at most one dim or the zero constant can appear in each result)"
                )
            if seen[expr.position]:
                raise VerifyException(
                    f'"{op.name}" requires a permutation_map that is a permutation '
                    "(found one dim used more than once)"
                )
            seen[expr.position] = True


@irdl_op_definition
class TransferReadOp(VectorTransferOperation):
    "Reads a supervector from memory into an SSA vector value."

    name = "vector.transfer_read"

    source = operand_def(TensorType | MemRefType)
    indices = var_operand_def(IndexType)
    padding = operand_def()
    mask = opt_operand_def(VectorType[I1])

    permutation_map = prop_def(AffineMapAttr)

    result = result_def(VectorType)

    irdl_options = [AttrSizedOperandSegments(as_property=True), ParsePropInAttrDict()]

    def __init__(
        self,
        source: SSAValue | Operation,
        indices: Sequence[SSAValue | Operation],
        padding: SSAValue | Operation,
        result_type: Attribute,
        in_bounds: ArrayAttr[BoolAttr],
        permutation_map: AffineMapAttr,
        mask: SSAValue | Operation | None = None,
    ):
        super().__init__(
            operands=[source, indices, padding, mask],
            result_types=[result_type],
            properties={"in_bounds": in_bounds, "permutation_map": permutation_map},
        )

    def print(self, printer: Printer):
        printer.print_string(" ", indent=0)
        printer.print_ssa_value(self.source)
        printer.print_string("[", indent=0)
        printer.print_list(self.indices, printer.print_ssa_value)
        printer.print_string("], ", indent=0)
        printer.print_ssa_value(self.padding)
        if self.mask is not None:
            printer.print_string(", ", indent=0)
            printer.print_ssa_value(self.mask)
        self._print_attrs(printer)
        printer.print_string(" : ", indent=0)
        printer.print_attribute(self.source.type)
        printer.print_string(", ", indent=0)
        printer.print_attribute(self.result.type)

    @classmethod
    def parse(cls, parser: Parser) -> TransferReadOp:
        source = parser.parse_unresolved_operand()
        indices = parser.parse_comma_separated_list(
            Parser.Delimiter.SQUARE, parser.parse_operand
        )
        parser.parse_punctuation(",")
        padding = parser.parse_operand()
        if parser.parse_optional_punctuation(","):
            mask_start_pos = parser.pos
            mask = parser.parse_unresolved_operand()
            mask_end_pos = parser.pos
        else:
            mask_start_pos = None
            mask = None
            mask_end_pos = None
        attributes_dict = parser.parse_optional_attr_dict()

        types_pos = parser.pos
        parser.parse_punctuation(":")
        shaped_type = parser.parse_type()
        parser.parse_punctuation(",")
        vector_type = parser.parse_type()

        source = parser.resolve_operand(source, shaped_type)

        if not isa(shaped_type, MemRefType | TensorType):
            parser.raise_error(
                "requires memref or ranked tensor type", at_position=types_pos
            )

        if not isa(vector_type, VectorType):
            parser.raise_error("requires vector type", at_position=types_pos)

        mask, permutation_map, in_bounds = VectorTransferOperation.resolve_attrs(
            parser,
            attributes_dict,
            shaped_type,
            vector_type,
            mask_start_pos,
            mask_end_pos,
            mask,
            types_pos,
        )

        # Create and return the TransferReadOp
        return TransferReadOp(
            source=source,
            indices=indices,
            padding=padding,
            mask=mask,
            permutation_map=permutation_map,
            in_bounds=in_bounds,
            result_type=vector_type,
        )

    def verify_(self):
        assert isa(self.source.type, MemRefType | TensorType)
        assert isa(self.result.type, VectorType)
        if self.mask:
            assert isa(self.mask.type, VectorType[I1])
            mask_type = self.mask.type
        else:
            mask_type = None

        if len(self.indices) != self.source.type.get_num_dims():
            raise VerifyException("Expected an index for each memref/tensor dimension.")

        if mask_type:
            inferred_mask_type = VectorTransferOperation.infer_transfer_op_mask_type(
                self.result.type,
                self.permutation_map.data,
            )
        else:
            inferred_mask_type = VectorType(i1, [])

        VectorTransferOperation.verify_op(
            self,
            self.source.type,
            self.result.type,
            mask_type,
            inferred_mask_type,
            self.permutation_map.data,
            self.in_bounds,
        )

        if isa(self.source.type.element_type, VectorType):
            # TODO verify vector element type
            pass
        else:
            # source memref/tensor has scalar element type
            # TODO verify that padding type is a valid element_type for a vector
            if self.source.type.element_type != self.padding.type:
                raise VerifyException(
                    f'"{self.name}" requires formal padding and source of the same elemental type'
                )

        VectorTransferOperation.verify_permutation_map(
            self,
            self.permutation_map.data,
        )


@irdl_op_definition
class TransferWriteOp(VectorTransferOperation):
    name = "vector.transfer_write"

    vector = operand_def(VectorType)
    source = operand_def(TensorType | MemRefType)
    indices = var_operand_def(IndexType)
    mask = opt_operand_def(VectorType[I1])

    permutation_map = prop_def(AffineMapAttr)

    result = opt_result_def(TensorType)

    irdl_options = [AttrSizedOperandSegments(as_property=True), ParsePropInAttrDict()]

    def __init__(
        self,
        vector: SSAValue | Operation,
        source: SSAValue | Operation,
        indices: Sequence[SSAValue | Operation],
        in_bounds: ArrayAttr[BoolAttr],
        mask: SSAValue | Operation | None = None,
        permutation_map: AffineMapAttr | None = None,
        result_type: TensorType | None = None,
    ):
        super().__init__(
            operands=[vector, source, indices, mask],
            properties={"in_bounds": in_bounds, "permutation_map": permutation_map},
            result_types=[result_type],
        )

    def print(self, printer: Printer):
        printer.print_string(" ", indent=0)
        printer.print_operand(self.vector)
        printer.print_string(", ", indent=0)
        printer.print_operand(self.source)
        printer.print_string("[", indent=0)
        printer.print_list(self.indices, printer.print_operand)
        printer.print_string("]", indent=0)
        if self.mask is not None:
            printer.print_string(", ", indent=0)
            printer.print_ssa_value(self.mask)
        self._print_attrs(printer)
        printer.print_string(" : ", indent=0)
        printer.print_attribute(self.vector.type)
        printer.print_string(", ", indent=0)
        printer.print_attribute(self.source.type)

    @classmethod
    def parse(cls, parser: Parser) -> TransferWriteOp:
        vector = parser.parse_unresolved_operand()
        parser.parse_punctuation(",")
        source = parser.parse_unresolved_operand()
        indices = parser.parse_comma_separated_list(
            Parser.Delimiter.SQUARE, parser.parse_operand
        )
        if parser.parse_optional_punctuation(","):
            mask_start_pos = parser.pos
            mask = parser.parse_unresolved_operand()
            mask_end_pos = parser.pos
        else:
            mask_start_pos = None
            mask = None
            mask_end_pos = None
        attributes_dict = parser.parse_optional_attr_dict()

        types_pos = parser.pos
        parser.parse_punctuation(":")
        vector_type = parser.parse_type()
        parser.parse_punctuation(",")
        shaped_type = parser.parse_type()

        vector = parser.resolve_operand(vector, vector_type)
        source = parser.resolve_operand(source, shaped_type)

        if not isa(shaped_type, MemRefType | TensorType):
            parser.raise_error(
                "requires memref or ranked tensor type", at_position=types_pos
            )

        if not isa(vector_type, VectorType):
            parser.raise_error("requires vector type", at_position=types_pos)

        mask, permutation_map, in_bounds = VectorTransferOperation.resolve_attrs(
            parser,
            attributes_dict,
            shaped_type,
            vector_type,
            mask_start_pos,
            mask_end_pos,
            mask,
            types_pos,
        )

        # Create and return the TransferReadOp
        return TransferWriteOp(
            vector=vector,
            source=source,
            indices=indices,
            mask=mask,
            permutation_map=permutation_map,
            in_bounds=in_bounds,
            result_type=shaped_type if isinstance(shaped_type, TensorType) else None,
        )

    def verify_(self):
        assert isa(self.source.type, MemRefType | TensorType)
        assert isa(self.vector.type, VectorType)
        if self.mask:
            assert isa(self.mask.type, VectorType[I1])
            mask_type = self.mask.type
        else:
            mask_type = None

        if len(self.indices) != self.source.type.get_num_dims():
            raise VerifyException("Expected an index for each memref/tensor dimension.")

        if self.has_broadcast_dim():
            raise VerifyException(
                f'"{self.name}" should not have broadcast dimensions.'
            )

        if mask_type:
            inferred_mask_type = VectorTransferOperation.infer_transfer_op_mask_type(
                self.vector.type,
                self.permutation_map.data,
            )
        else:
            inferred_mask_type = VectorType(i1, [])

        VectorTransferOperation.verify_op(
            self,
            self.source.type,
            self.vector.type,
            mask_type,
            inferred_mask_type,
            self.permutation_map.data,
            self.in_bounds,
        )

        VectorTransferOperation.verify_permutation_map(
            self,
            self.permutation_map.data,
        )


Vector = Dialect(
    "vector",
    [
        LoadOp,
        StoreOp,
        BroadcastOp,
        FMAOp,
        MaskedLoadOp,
        MaskedStoreOp,
        PrintOp,
        CreateMaskOp,
        ExtractOp,
        ExtractElementOp,
        InsertOp,
        InsertElementOp,
        TransferReadOp,
        TransferWriteOp,
    ],
    [],
)



import ast
from dataclasses import dataclass
from typing import Any

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.frontend.pyast.code_generation import CodeGeneration
from xdsl.frontend.pyast.utils.type_conversion import (
    FunctionRegistry,
    TypeConverter,
    TypeRegistry,
)
from xdsl.passes import PassPipeline


@dataclass
class PyASTBuilder:
    """Builder for xDSL modules from aspects of a Python function."""

    type_registry: TypeRegistry
    """Mappings between source code and IR type."""

    function_registry: FunctionRegistry
    """Mappings between functions and their operation types."""

    file: str | None
    """The file path of the function being built."""

    globals: dict[str, Any]
    """Global information for the function being built, including all the imports."""

    function_ast: ast.FunctionDef
    """The AST tree for the function being built."""

    build_context: Context
    """The xDSL context to use when applying transformations to the built module."""

    post_transforms: PassPipeline
    """An ordered list of passes and callbacks to apply to the built module."""

    def build(self) -> ModuleOp:
        """Build a module from the builder state."""
        # Convert the Python AST into xDSL IR objects
        type_converter = TypeConverter(
            self.globals,
            self.type_registry,
            self.function_registry,
        )
        module = CodeGeneration.run_with_type_converter(
            type_converter,
            self.function_ast,
            self.file,
        )
        module.verify()

        self.post_transforms.apply(self.build_context.clone(), module)
        return module



from __future__ import annotations

from itertools import islice
from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from xdsl.dialects.builtin import ModuleOp


class OpSelector(NamedTuple):
    """
    A specifier of an operation in a module.
    Useful when a stable reference is needed across copies of a module.
    """

    idx: int
    """
    The index of the operation in a module, when walking top to bottom as printed in the
    IR.
    """
    op_name: str
    """
    The name of the expected operation, used to check that the op found is the expected
    one.
    """

    def get_op(self, module_op: ModuleOp):
        """
        Returns the matching op, raising IndexError if the index is out of bounds and
        ValueError if the name does not match.
        """
        op = next(islice(module_op.walk(), self.idx, None), None)
        if op is None:
            raise IndexError(f"Matching index {self.idx} out of range.")
        if op.name != self.op_name:
            raise ValueError(
                f"Unexpected op {op.name} at index {self.idx}, expected {self.op_name}."
            )
        return op



from __future__ import annotations

from typing import Any, cast

from xdsl.dialects.builtin import TensorType, VectorType
from xdsl.interpreter import (
    Interpreter,
    InterpreterFunctions,
    PythonValues,
    ReturnedValues,
    TerminatorValue,
    impl,
    impl_callable,
    impl_terminator,
    register_impls,
)
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.utils.ptr import TypedPtr

from .dialects import toy as toy


@register_impls
class ToyFunctions(InterpreterFunctions):
    @impl(toy.PrintOp)
    def run_print(
        self, interpreter: Interpreter, op: toy.PrintOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        interpreter.print(f"{args[0]}")
        return ()

    @impl(toy.ConstantOp)
    def run_const(
        self, interpreter: Interpreter, op: toy.ConstantOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        assert not len(args)
        data = op.value.get_values()
        shape = list(op.value.get_shape())
        result = ShapedArray(TypedPtr.new_float64(data), shape)
        return (result,)

    @impl(toy.ReshapeOp)
    def run_reshape(
        self, interpreter: Interpreter, op: toy.ReshapeOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        (arg,) = args
        assert isinstance(arg, ShapedArray)
        arg = cast(ShapedArray[float], arg)
        result_type = op.results[0].type
        assert isinstance(result_type, VectorType | TensorType)
        new_shape = list(result_type.get_shape())

        return (ShapedArray(arg.data_ptr, new_shape),)

    @impl(toy.AddOp)
    def run_add(
        self, interpreter: Interpreter, op: toy.AddOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        lhs, rhs = args
        assert isinstance(lhs, ShapedArray)
        assert isinstance(rhs, ShapedArray)
        lhs = cast(ShapedArray[float], lhs)
        rhs = cast(ShapedArray[float], rhs)
        assert lhs.shape == rhs.shape

        return (
            ShapedArray(
                TypedPtr.new_float64([l + r for l, r in zip(lhs.data, rhs.data)]),
                lhs.shape,
            ),
        )

    @impl(toy.MulOp)
    def run_mul(
        self, interpreter: Interpreter, op: toy.MulOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        lhs, rhs = args
        assert isinstance(lhs, ShapedArray)
        assert isinstance(rhs, ShapedArray)
        lhs = cast(ShapedArray[float], lhs)
        rhs = cast(ShapedArray[float], rhs)
        assert lhs.shape == rhs.shape

        return (
            ShapedArray(
                TypedPtr.new_float64([l * r for l, r in zip(lhs.data, rhs.data)]),
                lhs.shape,
            ),
        )

    @impl_terminator(toy.ReturnOp)
    def run_return(
        self, interpreter: Interpreter, op: toy.ReturnOp, args: tuple[Any, ...]
    ) -> tuple[TerminatorValue, PythonValues]:
        assert len(args) < 2
        return ReturnedValues(args), ()

    @impl(toy.GenericCallOp)
    def run_generic_call(
        self, interpreter: Interpreter, op: toy.GenericCallOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        return interpreter.call_op(op.callee.string_value(), args)

    @impl(toy.TransposeOp)
    def run_transpose(
        self, interpreter: Interpreter, op: toy.TransposeOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        (arg,) = args
        assert isinstance(arg, ShapedArray)
        arg = cast(ShapedArray[float], arg)
        assert len(arg.shape) == 2

        result = arg.transposed(0, 1)

        return (result,)

    @impl(toy.CastOp)
    def run_cast(
        self, interpreter: Interpreter, op: toy.CastOp, args: tuple[Any, ...]
    ) -> tuple[Any, ...]:
        return args

    @impl_callable(toy.FuncOp)
    def call_func(
        self, interpreter: Interpreter, op: toy.FuncOp, args: tuple[Any, ...]
    ):
        return interpreter.run_ssacfg_region(op.body, args, op.sym_name.data)



#!/usr/bin/env python3
"""Benchmarks for the pipeline stages of the xDSL implementation.

This should live in its own file to avoid clobbering other benchmarks by
already having imported dependencies.
"""

import importlib

import xdsl
import xdsl.dialects.affine
import xdsl.dialects.arith
import xdsl.dialects.builtin
import xdsl.dialects.cf
import xdsl.dialects.test
import xdsl.interpreters.affine
import xdsl.interpreters.arith
import xdsl.interpreters.builtin
import xdsl.interpreters.cf
import xdsl.interpreters.func
import xdsl.interpreters.linalg
import xdsl.interpreters.memref
import xdsl.interpreters.memref_stream
import xdsl.interpreters.ml_program
import xdsl.interpreters.pdl
import xdsl.interpreters.printf
import xdsl.interpreters.riscv
import xdsl.interpreters.riscv_cf
import xdsl.interpreters.riscv_debug
import xdsl.interpreters.riscv_func
import xdsl.interpreters.riscv_libc
import xdsl.interpreters.riscv_scf
import xdsl.interpreters.riscv_snitch
import xdsl.interpreters.scf
import xdsl.interpreters.tensor


class ImportXDSL:
    """Benchmark importing xDSL."""

    def time_import_xdsl(self) -> None:
        """Time importing xDSL using the default asv mechanism."""
        importlib.reload(xdsl)


class ImportDialects:
    """Benchmark loading dialects in xDSL.

    Note that this must be done with `importlib.reload` rather than just
    directly importing with `from xdsl.dialects.arith import Arith` to avoid
    tests interacting with each other.
    """

    def ignore_time_affine_load(self) -> None:
        """Time loading the `arith` dialect."""
        importlib.reload(xdsl.dialects.affine)

    def ignore_time_arith_load(self) -> None:
        """Time loading the `arith` dialect."""
        importlib.reload(xdsl.dialects.arith)

    def ignore_time_builtin_load(self) -> None:
        """Time loading the `builtin` dialect."""
        importlib.reload(xdsl.dialects.builtin)

    def ignore_time_cf_load(self) -> None:
        """Time loading the `cf` dialect."""
        importlib.reload(xdsl.dialects.cf)

    def ignore_time_test_load(self) -> None:
        """Time loading the `test` dialect."""
        importlib.reload(xdsl.dialects.test)

    def time_all_constant_load(self) -> None:
        """Time all dialects used by the constant folding workload."""
        self.ignore_time_affine_load()
        self.ignore_time_arith_load()
        self.ignore_time_cf_load()


class ImportInterpreters:
    """Benchmark loading interpreters in xDSL."""

    def time_all_constant_load(self) -> None:
        """Time all interpreters used by the constant folding workload."""
        importlib.reload(xdsl.dialects.test)
        importlib.reload(xdsl.interpreters.affine)
        importlib.reload(xdsl.interpreters.tensor)
        importlib.reload(xdsl.interpreters.scf)
        importlib.reload(xdsl.interpreters.riscv_snitch)
        importlib.reload(xdsl.interpreters.riscv_scf)
        importlib.reload(xdsl.interpreters.riscv_libc)
        importlib.reload(xdsl.interpreters.riscv_func)
        importlib.reload(xdsl.interpreters.riscv_debug)
        importlib.reload(xdsl.interpreters.riscv_cf)
        importlib.reload(xdsl.interpreters.riscv)
        importlib.reload(xdsl.interpreters.printf)
        importlib.reload(xdsl.interpreters.pdl)
        importlib.reload(xdsl.interpreters.ml_program)
        importlib.reload(xdsl.interpreters.memref_stream)
        importlib.reload(xdsl.interpreters.memref)
        importlib.reload(xdsl.interpreters.linalg)
        importlib.reload(xdsl.interpreters.func)
        importlib.reload(xdsl.interpreters.cf)
        importlib.reload(xdsl.interpreters.builtin)
        importlib.reload(xdsl.interpreters.arith)


if __name__ == "__main__":
    from bench_utils import Benchmark, profile

    XDSL = ImportXDSL()
    DIALECTS = ImportDialects()
    INTERPRETERS = ImportInterpreters()
    profile(
        {
            "xDSL": Benchmark(XDSL.time_import_xdsl),
            "Dialects.affine_load": Benchmark(DIALECTS.ignore_time_affine_load),
            "Dialects.arith_load": Benchmark(DIALECTS.ignore_time_arith_load),
            "Dialects.builtin_load": Benchmark(DIALECTS.ignore_time_builtin_load),
            "Dialects.cf_load": Benchmark(DIALECTS.ignore_time_cf_load),
            "Dialects.test_load": Benchmark(DIALECTS.ignore_time_test_load),
            "Dialects.all_constant_load": Benchmark(DIALECTS.time_all_constant_load),
            "Interpreters.all_constant_load": Benchmark(
                INTERPRETERS.time_all_constant_load
            ),
        }
    )



import pytest

from xdsl.dialects.arith import ConstantOp, FloatingPointLikeBinaryOperation
from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
    FloatAttr,
    TensorType,
    VectorType,
    f32,
    i32,
)
from xdsl.dialects.math import (
    AbsFOp,
    AbsIOp,
    AcoshOp,
    AcosOp,
    AsinhOp,
    AsinOp,
    Atan2Op,
    AtanhOp,
    AtanOp,
    CbrtOp,
    CeilOp,
    CopySignOp,
    CoshOp,
    CosOp,
    CountLeadingZerosOp,
    CountTrailingZerosOp,
    CtPopOp,
    ErfOp,
    Exp2Op,
    ExpM1Op,
    ExpOp,
    FloorOp,
    FmaOp,
    FPowIOp,
    IPowIOp,
    Log1pOp,
    Log2Op,
    Log10Op,
    LogOp,
    PowFOp,
    RoundEvenOp,
    RoundOp,
    RsqrtOp,
    SinhOp,
    SinOp,
    SqrtOp,
    TanhOp,
    TanOp,
    TruncOp,
)
from xdsl.dialects.test import TestOp
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import create_ssa_value


class Test_float_math_binary_construction:
    operand_type = f32
    a = ConstantOp(FloatAttr(1.1, operand_type))
    b = ConstantOp(FloatAttr(2.2, operand_type))

    f32_vector_type = VectorType(f32, [3])

    lhs_vector_ssa_value = create_ssa_value(f32_vector_type)
    rhs_vector_ssa_value = create_ssa_value(f32_vector_type)

    lhs_vector = TestOp([lhs_vector_ssa_value], result_types=[f32_vector_type])
    rhs_vector = TestOp([rhs_vector_ssa_value], result_types=[f32_vector_type])

    f32_tensor_type = DenseIntOrFPElementsAttr.from_list(TensorType(f32, []), [5.5])
    lhs_tensor_ssa_value = create_ssa_value(f32_tensor_type)
    rhs_tensor_ssa_value = create_ssa_value(f32_tensor_type)

    lhs_tensor = TestOp([lhs_tensor_ssa_value], result_types=[f32_tensor_type])
    rhs_tensor = TestOp([rhs_tensor_ssa_value], result_types=[f32_tensor_type])

    @pytest.mark.parametrize(
        "OpClass",
        [
            Atan2Op,
            CopySignOp,
            IPowIOp,
            PowFOp,
        ],
    )
    def test_float_binary_ops_constant_math_init(
        self,
        OpClass: type[FloatingPointLikeBinaryOperation],
    ):
        op = OpClass(self.a, self.b)
        assert isinstance(op, OpClass)
        assert op.lhs.owner is self.a
        assert op.rhs.owner is self.b
        assert op.result.type == self.operand_type

    @pytest.mark.parametrize(
        "OpClass",
        [
            Atan2Op,
            CopySignOp,
            IPowIOp,
            PowFOp,
        ],
    )
    def test_float_binary_vector_ops_init(
        self,
        OpClass: type[FloatingPointLikeBinaryOperation],
    ):
        op = OpClass(self.lhs_vector, self.rhs_vector)
        assert isinstance(op, OpClass)
        assert op.lhs.owner is self.lhs_vector
        assert op.rhs.owner is self.rhs_vector
        assert op.result.type == self.f32_vector_type

    @pytest.mark.parametrize(
        "OpClass",
        [
            Atan2Op,
            CopySignOp,
            IPowIOp,
            PowFOp,
        ],
    )
    def test_float_binary_ops_tensor_math_init(
        self,
        OpClass: type[FloatingPointLikeBinaryOperation],
    ):
        op = OpClass(self.lhs_tensor, self.rhs_tensor)
        assert isinstance(op, OpClass)
        assert op.lhs.owner is self.lhs_tensor
        assert op.rhs.owner is self.rhs_tensor
        assert op.result.type == self.f32_tensor_type


class Test_float_math_unary_constructions:
    operand_type = f32
    a = ConstantOp(FloatAttr(1, operand_type))

    f32_vector_type = VectorType(f32, [3])

    test_vector_ssa = create_ssa_value(f32_vector_type)
    test_vec = TestOp([test_vector_ssa], result_types=[f32_vector_type])

    f32_tensor_type = DenseIntOrFPElementsAttr.from_list(TensorType(f32, []), [5.5])
    test_tensor_ssa_value = create_ssa_value(f32_tensor_type)
    test_tensor = TestOp([test_tensor_ssa_value], result_types=[f32_tensor_type])

    @pytest.mark.parametrize(
        "OpClass",
        [
            AbsFOp,
            AcosOp,
            AcoshOp,
            AsinOp,
            AsinhOp,
            AtanOp,
            AtanhOp,
            CbrtOp,
            CeilOp,
            CosOp,
            CoshOp,
            ErfOp,
            Exp2Op,
            ExpM1Op,
            ExpOp,
            FloorOp,
            Log10Op,
            Log1pOp,
            Log2Op,
            LogOp,
            RoundEvenOp,
            RoundOp,
            RsqrtOp,
            SinOp,
            SinhOp,
            SqrtOp,
            TanOp,
            TanhOp,
            TruncOp,
        ],
    )
    def test_float_math_constant_ops_init(
        self,
        OpClass: type,
    ):
        op = OpClass(self.a)
        assert op.result.type == f32
        assert op.operand.type == f32
        assert op.operand.owner == self.a
        assert op.result.type == self.operand_type

    @pytest.mark.parametrize(
        "OpClass",
        [
            AbsFOp,
            AcosOp,
            AcoshOp,
            AsinOp,
            AsinhOp,
            AtanOp,
            AtanhOp,
            CbrtOp,
            CeilOp,
            CosOp,
            CoshOp,
            ErfOp,
            Exp2Op,
            ExpM1Op,
            ExpOp,
            FloorOp,
            Log10Op,
            Log1pOp,
            Log2Op,
            LogOp,
            RoundEvenOp,
            RoundOp,
            RsqrtOp,
            SinOp,
            SinhOp,
            SqrtOp,
            TanOp,
            TanhOp,
            TruncOp,
        ],
    )
    def test_float_math_ops_vector_init(self, OpClass: type):
        op = OpClass(self.test_vec)
        assert op.result.type == self.f32_vector_type
        assert op.operand.type == self.f32_vector_type
        assert op.operand.owner == self.test_vec
        assert op.result.type == self.f32_vector_type

    @pytest.mark.parametrize(
        "OpClass",
        [
            AbsFOp,
            AcosOp,
            AcoshOp,
            AsinOp,
            AsinhOp,
            AtanOp,
            AtanhOp,
            CbrtOp,
            CeilOp,
            CosOp,
            CoshOp,
            ErfOp,
            Exp2Op,
            ExpM1Op,
            ExpOp,
            FloorOp,
            Log10Op,
            Log1pOp,
            Log2Op,
            LogOp,
            RoundEvenOp,
            RoundOp,
            RsqrtOp,
            SinOp,
            SinhOp,
            SqrtOp,
            TanOp,
            TanhOp,
            TruncOp,
        ],
    )
    def test_float_math_ops_tensor_init(
        self,
        OpClass: type,
    ):
        op = OpClass(self.test_tensor)
        assert op.result.type == self.f32_tensor_type
        assert op.operand.type == self.f32_tensor_type
        assert op.operand.owner == self.test_tensor
        assert op.result.type == self.f32_tensor_type


class Test_fpowi:
    a = ConstantOp(FloatAttr(2.2, f32))
    b = ConstantOp.from_int_and_width(0, 32)

    f32_vector_type = VectorType(f32, [3])

    a_vector_ssa_value = create_ssa_value(f32_vector_type)
    a_vector = TestOp([a_vector_ssa_value], result_types=[f32_vector_type])

    f32_tensor_type = DenseIntOrFPElementsAttr.from_list(TensorType(f32, []), [5.5])
    a_tensor_ssa_value = create_ssa_value(f32_tensor_type)
    a_tensor = TestOp([a_tensor_ssa_value], result_types=[f32_tensor_type])

    def test_fpowi_constant_construction(self):
        op = FPowIOp(self.a, self.b)
        assert op.result.type == f32
        assert op.lhs.owner is self.a
        assert op.rhs.owner is self.b
        assert op.result.type == f32

    def test_fpowi_vector_construction(self):
        op = FPowIOp(self.a_vector, self.b)
        assert op.result.type == self.f32_vector_type
        assert op.lhs.owner is self.a_vector
        assert op.rhs.owner is self.b
        assert op.result.type == self.f32_vector_type

    def test_fpowi_tensor_construction(self):
        op = FPowIOp(self.a_tensor, self.b)
        assert op.result.type == self.f32_tensor_type
        assert op.lhs.owner is self.a_tensor
        assert op.rhs.owner is self.b
        assert op.result.type == self.f32_tensor_type


class Test_fma:
    a = ConstantOp(FloatAttr(1.1, f32))
    b = ConstantOp(FloatAttr(2.2, f32))
    c = ConstantOp(FloatAttr(3.3, f32))

    f32_vector_type = VectorType(f32, [3])
    test_vector_ssa = create_ssa_value(f32_vector_type)
    a_vec = TestOp([test_vector_ssa], result_types=[f32_vector_type])
    b_vec = TestOp([test_vector_ssa], result_types=[f32_vector_type])
    c_vec = TestOp([test_vector_ssa], result_types=[f32_vector_type])

    f32_tensor_type = DenseIntOrFPElementsAttr.from_list(TensorType(f32, []), [5.5])
    test_tensor_ssa = create_ssa_value(f32_vector_type)
    a_tensor = TestOp([test_tensor_ssa], result_types=[f32_tensor_type])
    b_tensor = TestOp([test_tensor_ssa], result_types=[f32_tensor_type])
    c_tensor = TestOp([test_tensor_ssa], result_types=[f32_tensor_type])

    def test_fma_construction(self):
        op = FmaOp(self.a, self.b, self.c)
        assert op.result.type == f32
        assert op.a.owner is self.a
        assert op.b.owner is self.b
        assert op.c.owner is self.c
        assert op.result.type == f32

    def test_fma_construction_vector(self):
        op = FmaOp(self.a_vec, self.b_vec, self.c_vec)
        assert op.result.type == self.f32_vector_type
        assert op.a.owner is self.a_vec
        assert op.b.owner is self.b_vec
        assert op.c.owner is self.c_vec

    def test_fma_construction_tensor(self):
        op = FmaOp(self.a_tensor, self.b_tensor, self.c_tensor)
        assert op.result.type == self.f32_tensor_type
        assert op.a.owner is self.a_tensor
        assert op.b.owner is self.b_tensor
        assert op.c.owner is self.c_tensor


class Test_int_math_unary_constructions:
    operand_type = i32
    a = ConstantOp.from_int_and_width(0, 32)

    i32_vector_type = VectorType(i32, [1])
    test_vector_ssa = create_ssa_value(i32_vector_type)
    test_vec = TestOp([test_vector_ssa], result_types=[i32_vector_type])

    i32_tensor_type = DenseIntOrFPElementsAttr.from_list(TensorType(i32, []), [5])
    test_tensor_ssa_value = create_ssa_value(i32_tensor_type)
    test_tensor = TestOp([test_tensor_ssa_value], result_types=[i32_tensor_type])

    @pytest.mark.parametrize(
        "OpClass",
        [
            AbsIOp,
            CountLeadingZerosOp,
            CountTrailingZerosOp,
            CtPopOp,
        ],
    )
    def test_int_math_ops_init(
        self,
        OpClass: type,
    ):
        op = OpClass(self.a)
        assert op.result.type == i32
        assert op.operand.type == i32
        assert op.operand.owner == self.a
        assert op.result.type == self.operand_type

    @pytest.mark.parametrize(
        "OpClass",
        [
            AbsIOp,
            CountLeadingZerosOp,
            CountTrailingZerosOp,
            CtPopOp,
        ],
    )
    def test_int_math_ops_vec_init(
        self,
        OpClass: type,
    ):
        op = OpClass(self.test_vec)
        assert op.result.type == self.i32_vector_type
        assert op.operand.type == self.i32_vector_type
        assert op.operand.owner == self.test_vec
        assert op.result.type == self.i32_vector_type

    @pytest.mark.parametrize(
        "OpClass",
        [
            AbsIOp,
            CountLeadingZerosOp,
            CountTrailingZerosOp,
            CtPopOp,
        ],
    )
    def test_int_math_ops_tensor_init(
        self,
        OpClass: type,
    ):
        op = OpClass(self.test_tensor)
        assert op.result.type == self.i32_tensor_type
        assert op.operand.type == self.i32_tensor_type
        assert op.operand.owner == self.test_tensor
        assert op.result.type == self.i32_tensor_type


class Test_Trunci:
    operand_type = i32
    a = ConstantOp.from_int_and_width(0, 32)

    i32_vector_type = VectorType(i32, [1])
    test_vector_ssa = create_ssa_value(i32_vector_type)
    test_vec = TestOp([test_vector_ssa], result_types=[i32_vector_type])

    i32_tensor_type = DenseIntOrFPElementsAttr.from_list(TensorType(i32, []), [5])
    test_tensor_ssa_value = create_ssa_value(i32_tensor_type)
    test_tensor = TestOp([test_tensor_ssa_value], result_types=[i32_tensor_type])

    def test_trunci_incorrect_bitwidth(self):
        with pytest.raises(VerifyException):
            TruncOp(self.a).verify()
        with pytest.raises(VerifyException):
            TruncOp(self.test_vec).verify()
        with pytest.raises(VerifyException):
            TruncOp(self.test_tensor).verify()



import importlib
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import (
    Any,
    cast,
)

from xdsl.frontend.pyast.utils.exceptions import (
    FrontendProgramException,
)
from xdsl.ir import Attribute, Operation, SSAValue, TypeAttribute


class FunctionRegistry:
    """A mapping between Python callables and IR operation types."""

    def __init__(self):
        """Instantiate the function registry."""
        self._mapping: dict[Callable[..., Any], Callable[..., Operation]] = {}

    def insert(
        self, callable: Callable[..., Any], ir_constructor: Callable[..., Operation]
    ) -> None:
        """Insert a relation between a Python callable and an IR operation constructor."""
        if callable in self._mapping:
            raise FrontendProgramException(
                f"Cannot re-register function '{callable.__qualname__}'"
            )
        self._mapping[callable] = ir_constructor

    def get_operation_constructor(
        self, callable: Callable[..., Any]
    ) -> Callable[..., Operation] | None:
        """Get the IR operation constructor from a Python callable."""
        return self._mapping.get(callable, None)

    def resolve_operation(
        self,
        module_name: str,
        method_name: str,
        args: tuple[SSAValue[Attribute], ...] = tuple(),
        kwargs: dict[str, SSAValue[Attribute]] = dict(),
    ) -> Operation | None:
        """Get a concrete IR operation from a method name and its arguments."""
        # Start at the module, and walk till a method leaf is found
        method = importlib.import_module(module_name)
        for attr in method_name.split("."):
            method = getattr(method, attr, None)
        if method is None:
            return None
        assert callable(method), "Guaranteed by type signature of registration method"
        if (
            operation_constructor := self.get_operation_constructor(method)
        ) is not None:
            return operation_constructor(*args, **kwargs)
        return None


class TypeRegistry:
    """A mapping between Python type annotations and IR type attributes."""

    def __init__(self):
        """Instantiate the function registry."""
        self._mapping: dict[type, TypeAttribute] = {}

    def insert(
        self,
        annotation: type,
        attribute: TypeAttribute,
    ) -> None:
        """Insert a relation between a Python type annotation and an IR type attribute."""
        if annotation in self._mapping:
            raise FrontendProgramException(
                f"Cannot re-register type name '{annotation.__qualname__}'"
            )
        if attribute in self._mapping.values():
            raise FrontendProgramException(
                f"Cannot register multiple source types for IR type '{attribute}'"
            )
        self._mapping[annotation] = attribute

    def get_annotation(self, attribute: TypeAttribute) -> type | None:
        """Get the Python type annotation from an IR type attribute.

        This supports many-to-one mappings by resolving greedily on the first
        mapping inserted.
        """
        for key, value in self._mapping.items():
            if value == attribute:
                return key
        return None

    def resolve_attribute(
        self, annotation_name: str, globals: dict[str, Any]
    ) -> TypeAttribute | None:
        """Get an IR type attribute from a string annotation."""
        annotation = cast(
            type,
            eval(annotation_name, globals, None),
        )
        return self._mapping.get(annotation, None)


@dataclass
class TypeConverter:
    """Responsible for conversion of Python type hints to xDSL types."""

    globals: dict[str, Any] = field(default_factory=dict[str, Any])
    """
    Stores all globals in the current Python program, including imports. This is
    useful because we can lookup a class which corresponds to the type
    annotation without explicitly constructing it.
    """

    type_registry: TypeRegistry = field(default_factory=TypeRegistry)
    """Mappings between source code and ir type, indexed by name."""

    function_registry: FunctionRegistry = field(default_factory=FunctionRegistry)
    """Mappings between methods on objects and their operations."""



from dataclasses import dataclass
from typing import cast

from xdsl.context import Context
from xdsl.dialects import arith, ptr
from xdsl.dialects.builtin import FixedBitwidthType, IndexType, ModuleOp
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.exceptions import DiagnosticException


@dataclass
class ConvertTypeOffsetOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ptr.TypeOffsetOp, rewriter: PatternRewriter, /):
        if not issubclass(type(op.elem_type), FixedBitwidthType):
            raise DiagnosticException(
                "Type offset is currently only supported for fixed size types"
            )
        elem_type = cast(FixedBitwidthType, op.elem_type)
        rewriter.replace_matched_op(
            arith.ConstantOp.from_int_and_width(elem_type.size, IndexType())
        )


class ConvertPtrTypeOffsetsPass(ModulePass):
    name = "convert-ptr-type-offsets"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(
            ConvertTypeOffsetOp(),
        ).rewrite_module(op)



"""
This file contains the data structures necessary for the parsing and printing
of the MLIR declarative assembly format defined at
https://mlir.llvm.org/docs/DefiningDialects/Operations/#declarative-assembly-format .
"""

from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Any, ClassVar, Literal, cast

from typing_extensions import TypeVar

from xdsl.dialects.builtin import (
    AnyFloat,
    BytesAttr,
    DenseArrayBase,
    IntegerType,
    StringAttr,
    UnitAttr,
)
from xdsl.ir import (
    Attribute,
    Data,
    ParametrizedAttribute,
    Region,
    SSAValue,
    TypedAttribute,
)
from xdsl.irdl import (
    ConstraintContext,
    IRDLOperation,
    IRDLOperationInvT,
    OpDef,
    OptionalDef,
    Successor,
    VarOperand,
    is_const_classvar,
)
from xdsl.parser import Parser, UnresolvedOperand
from xdsl.printer import Printer
from xdsl.utils.exceptions import PyRDLError
from xdsl.utils.hints import isa
from xdsl.utils.mlir_lexer import PunctuationSpelling


@dataclass
class ParsingState:
    """
    State carried during the parsing of an operation using the declarative assembly
    format.
    It contains the elements that have already been parsed.
    """

    operands: list[None | Sequence[UnresolvedOperand]]
    operand_types: list[None | Sequence[Attribute]]
    result_types: list[None | Sequence[Attribute]]
    regions: list[None | Sequence[Region]]
    successors: list[None | Sequence[Successor]]
    attributes: dict[str, Attribute]
    properties: dict[str, Attribute]
    context: ConstraintContext

    def __init__(self, op_def: OpDef):
        self.operands = [None] * len(op_def.operands)
        self.operand_types = [None] * len(op_def.operands)
        self.result_types = [None] * len(op_def.results)
        self.regions = [None] * len(op_def.regions)
        self.successors = [None] * len(op_def.successors)
        self.attributes = {}
        self.properties = {}
        self.context = ConstraintContext()


@dataclass
class PrintingState:
    """
    State carried during the printing of an operation using the declarative assembly
    format.
    It contains information on the last token, to know if a space should be emitted.
    """

    last_was_punctuation: bool = field(default=False)
    """Was the last element parsed a punctuation."""
    should_emit_space: bool = field(default=True)
    """
    Should the printer emit a space before the next element.
    Depending on the directive, the space might not be printed
    (for instance for some punctuations).
    """

    def print_whitespace(self, printer: Printer):
        """
        Handles whitespace printing for the majority of format directives.
        """
        if self.should_emit_space or not self.last_was_punctuation:
            printer.print_string(" ")
        self.should_emit_space = True
        self.last_was_punctuation = False


@dataclass(frozen=True)
class FormatProgram:
    """
    The toplevel data structure of a declarative assembly format program.
    It is used to parse and print an operation.
    """

    stmts: tuple[FormatDirective, ...]
    """The statements composing the program. They are executed in order."""

    @staticmethod
    def from_str(input: str, op_def: OpDef) -> FormatProgram:
        """
        Create the assembly format data program from its string representation.
        This might raise a ParseError exception if the string is invalid.
        """
        from xdsl.irdl.declarative_assembly_format_parser import FormatParser

        return FormatParser(input, op_def).parse_format()

    def parse(
        self, parser: Parser, op_type: type[IRDLOperationInvT]
    ) -> IRDLOperationInvT:
        """
        Parse the operation with this format.
        The given operation type is expected to be the operation type represented by
        the operation definition passed to the FormatParser that created this
        FormatProgram.
        """
        # Parse elements one by one
        op_def = op_type.get_irdl_definition()
        state = ParsingState(op_def)
        for stmt in self.stmts:
            stmt.parse(parser, state)

        # Get constraint variables from the parsed operand and result types
        self.resolve_constraint_variables(state, op_def)

        # Infer operand types that should be inferred
        assert None not in state.operands
        unresolved_operands = cast(list[Sequence[UnresolvedOperand]], state.operands)
        operand_types = self.resolve_operand_types(state, op_def)

        # Infer result types that should be inferred
        result_types = self.resolve_result_types(state, op_def)

        # Resolve all operands
        operands = tuple(
            parser.resolve_operands(uo, ot, parser.pos)
            for uo, ot in zip(unresolved_operands, operand_types, strict=True)
        )

        return op_type.build(
            result_types=result_types,
            operands=operands,
            attributes=state.attributes,
            properties=state.properties,
            regions=state.regions,
            successors=state.successors,
        )

    def resolve_constraint_variables(self, state: ParsingState, op_def: OpDef):
        """
        Runs verification on the parsed parts of the operation, adding the resolved value
        of each constraint variable to the `ConstraintContext` `state.context`.
        """
        ctx = state.context

        for operand, operand_type, (_, operand_def) in zip(
            state.operands, state.operand_types, op_def.operands, strict=True
        ):
            length = len(operand) if isinstance(operand, Sequence) else 1
            operand_def.constr.verify_length(length, ctx)
            if operand_type is None:
                continue
            if isinstance(operand_type, Attribute):
                operand_type = (operand_type,)
            operand_def.constr.verify(operand_type, ctx)

        for result_type, (_, result_def) in zip(
            state.result_types, op_def.results, strict=True
        ):
            if result_type is None:
                continue
            if isinstance(result_type, Attribute):
                result_type = (result_type,)
            result_def.constr.verify(result_type, ctx)

        for prop_name, prop_def in op_def.properties.items():
            if isinstance(prop_def, OptionalDef) and prop_def.default_value is None:
                continue
            attr = state.properties.get(prop_name, prop_def.default_value)
            if attr is None:
                continue
            prop_def.constr.verify(attr, ctx)

        for attr_name, attr_def in op_def.attributes.items():
            if isinstance(attr_def, OptionalDef) and attr_def.default_value is None:
                continue
            attr = state.attributes.get(attr_name, attr_def.default_value)
            if attr is None:
                continue
            attr_def.constr.verify(attr, ctx)

    def resolve_operand_types(
        self, state: ParsingState, op_def: OpDef
    ) -> Sequence[Sequence[Attribute]]:
        """
        Use the inferred type resolutions to fill missing operand types from other parsed
        types.
        """
        return tuple(
            operand_def.constr.infer(
                state.context,
                length=len(cast(Sequence[Any], operand)),
            )
            if operand_type is None
            else operand_type
            for operand_type, operand, (_, operand_def) in zip(
                state.operand_types, state.operands, op_def.operands, strict=True
            )
        )

    def resolve_result_types(
        self, state: ParsingState, op_def: OpDef
    ) -> Sequence[Sequence[Attribute]]:
        """
        Use the inferred type resolutions to fill missing result types from other parsed
        types.
        """
        return tuple(
            result_def.constr.infer(
                state.context,
                length=None,
            )
            if result_type is None
            else result_type
            for result_type, (_, result_def) in zip(
                state.result_types, op_def.results, strict=True
            )
        )

    def print(self, printer: Printer, op: IRDLOperation) -> None:
        """
        Print the operation with this format.
        The given operation is expected to be defined using the operation definition
        passed to the FormatParser that created this FormatProgram.
        """
        state = PrintingState()
        for stmt in self.stmts:
            stmt.print(printer, state, op)


@dataclass(frozen=True)
class Directive(ABC):
    """An assembly format directive"""

    def is_present(self, op: IRDLOperation) -> bool:
        """
        Check if the directive is present in the input.
        """
        return True

    def is_anchorable(self) -> bool:
        """
        Can appear as an anchor in an optional group.
        """
        return False

    def is_variadic_like(self) -> bool:
        """
        Variadic-like format directives parse a comma separated list, and cannot be
        followed by `,` directive.
        """
        return False

    def is_optional_like(self) -> bool:
        """
        Directives that successfully parse the empty string.
        The first element in an optional group must be optional-like.
        """
        return self.is_variadic_like()


class FormatDirective(Directive, ABC):
    """A format directive for operation format."""

    @abstractmethod
    def parse(self, parser: Parser, state: ParsingState) -> bool:
        """
        Parses the directive, returning True if input was consumed.
        """
        ...

    @abstractmethod
    def print(
        self, printer: Printer, state: PrintingState, op: IRDLOperation
    ) -> None: ...

    def set_empty(self, state: ParsingState):
        """
        Set the appropriate field of the parsing state to be empty.
        Used when a variable appears in an optional group which is not parsed.
        """
        return


class CustomDirective(FormatDirective, ABC):
    """
    A user defined assembly format directive.
    A custom directive can have multiple parameters, whose types should be
    declared in the `parameters` field.
    """

    parameters: ClassVar[dict[str, type[FormatDirective]]]


CustomDirectiveInvT = TypeVar("CustomDirectiveInvT", bound=CustomDirective)


def irdl_custom_directive(cls: type[CustomDirectiveInvT]) -> type[CustomDirectiveInvT]:
    """Decorator used on custom directives to define the `parameters` class variable."""

    cls.parameters = {}
    param_types = inspect.get_annotations(cls, eval_str=True)
    for field_name, ty in param_types.items():
        if is_const_classvar(field_name, ty, PyRDLError):
            continue
        if not issubclass(ty, FormatDirective):
            raise PyRDLError(
                f"Custom directive {cls.__name__} has parameter {field_name} which is not a format directive."
            )
        cls.parameters[field_name] = ty
    return dataclass(frozen=True)(cls)


class TypeableDirective(Directive, ABC):
    """
    Directives which can be used to set or get types.
    """

    @abstractmethod
    def set_types(self, state: ParsingState, types: Sequence[Attribute]) -> str | None:
        """
        Set the types for this directive to the given sequence, possibly returning an error.
        """
        ...

    @abstractmethod
    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        """
        Parses types for the directive, returning True if input was consumed.
        """
        ...

    @abstractmethod
    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        """
        Parse exactly one type for the directive.
        """

    @abstractmethod
    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]: ...


@dataclass(frozen=True)
class TypeDirective(FormatDirective):
    """
    A directive which parses the type of a typeable directive, with format:
      type-directive ::= type(typeable-directive)
    """

    inner: TypeableDirective

    def set(self, state: ParsingState, types: Sequence[Attribute]):
        self.inner.set_types(state, types)

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        return self.inner.parse_types(parser, state)

    def get(self, op: IRDLOperation) -> Sequence[Attribute]:
        return self.inner.get_types(op)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        types = self.inner.get_types(op)
        if not types:
            return
        state.print_whitespace(printer)
        printer.print_list(types, printer.print_attribute)

    def is_present(self, op: IRDLOperation) -> bool:
        return self.inner.is_present(op)

    def is_anchorable(self) -> bool:
        return self.inner.is_anchorable()

    def is_variadic_like(self) -> bool:
        return self.inner.is_variadic_like()

    def is_optional_like(self) -> bool:
        return self.inner.is_optional_like()

    def set_empty(self, state: ParsingState):
        self.set(state, ())


@dataclass(frozen=True)
class VariableDirective(Directive, ABC):
    """
    A variable directive, with the following format:
      variable-directive ::= dollar-ident
    The directive will request a space to be printed after.
    """

    name: str
    """The variable name. This is only used for error message reporting."""
    index: int
    """Index of the variable(operand or result) definition."""


class VariadicVariable(VariableDirective, ABC):
    def is_present(self, op: IRDLOperation) -> bool:
        return bool(getattr(op, self.name))

    def is_anchorable(self) -> bool:
        return True

    def is_variadic_like(self) -> bool:
        return True


class OptionalVariable(VariableDirective, ABC):
    def is_present(self, op: IRDLOperation) -> bool:
        return getattr(op, self.name) is not None

    def is_anchorable(self) -> bool:
        return True

    def is_optional_like(self) -> bool:
        return True


@dataclass(frozen=True)
class AttrDictDirective(FormatDirective):
    """
    An attribute dictionary directive, with the following format:
       attr-dict-directive ::= attr-dict
       attr-dict-with-format-directive ::= `attributes` attr-dict
    The directive (with and without the keyword) will always print a space before, and
    will not request a space to be printed after.
    """

    with_keyword: bool
    """If this is set, the format starts with the `attributes` keyword."""

    reserved_attr_names: set[str]
    """
    The set of attributes that should not be printed.
    These attributes are printed in other places in the format, and thus would be
    printed twice otherwise.
    """

    expected_properties: set[str]
    """
    Properties that should be printed and parsed as part of this attr-dict.
    This is used to keep compatibility with MLIR which allows that.
    """

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        if self.with_keyword:
            res = parser.parse_optional_attr_dict_with_keyword()
            if res is None:
                res = {}
            else:
                res = dict(res.data)
        else:
            res = parser.parse_optional_attr_dict()
        defined_reserved_keys = self.reserved_attr_names & res.keys()
        if defined_reserved_keys:
            parser.raise_error(
                f"attributes {', '.join(defined_reserved_keys)} are defined in other parts of the "
                "assembly format, and thus should not be defined in the attribute "
                "dictionary."
            )

        props = tuple(k for k in res.keys() if k in self.expected_properties)
        for name in props:
            state.properties[name] = res.pop(name)
        state.attributes |= res
        return bool(res) or bool(props)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        if not op.attributes.keys().isdisjoint(self.expected_properties):
            raise ValueError(
                "Cannot print attributes and properties with the same name "
                "in a single dictionary"
            )
        op_def = op.get_irdl_definition()
        dictionary = op.attributes | {
            k: v for k, v in op.properties.items() if k in self.expected_properties
        }
        defs = {
            x: op_def.properties[x] for x in self.expected_properties
        } | op_def.attributes

        reserved_or_default = self.reserved_attr_names.union(
            name
            for name, d in defs.items()
            if d.default_value is not None and dictionary.get(name) == d.default_value
        )

        printed = printer.print_op_attributes(
            dictionary,
            reserved_attr_names=reserved_or_default,
            print_keyword=self.with_keyword,
        )

        if printed:
            state.last_was_punctuation = False
            state.should_emit_space = True

    def is_optional_like(self) -> bool:
        return True


class OperandDirective(FormatDirective, TypeableDirective, ABC):
    """
    Base class for operand directives to aid typechecking.
    """

    pass


@dataclass(frozen=True)
class OperandVariable(VariableDirective, OperandDirective):
    """
    An operand variable, with the following format:
      operand-directive ::= dollar-ident
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, operand: UnresolvedOperand):
        state.operands[self.index] = (operand,)

    def set_types(self, state: ParsingState, types: Sequence[Attribute]):
        state.operand_types[self.index] = types

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        operand = parser.parse_unresolved_operand()
        self.set(state, operand)
        return True

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        self.set_types(state, (parser.parse_type(),))
        return True

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        self.parse_types(parser, state)

    def get(self, op: IRDLOperation) -> SSAValue:
        return getattr(op, self.name)

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        return (self.get(op).type,)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        state.print_whitespace(printer)
        printer.print_ssa_value(self.get(op))


@dataclass(frozen=True)
class VariadicOperandVariable(VariadicVariable, OperandDirective):
    """
    A variadic operand variable, with the following format:
      operand-directive ::= ( percent-ident ( `,` percent-id )* )?
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, operands: Sequence[UnresolvedOperand]):
        state.operands[self.index] = operands

    def set_types(self, state: ParsingState, types: Sequence[Attribute]):
        state.operand_types[self.index] = types

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        operands = parser.parse_optional_undelimited_comma_separated_list(
            parser.parse_optional_unresolved_operand, parser.parse_unresolved_operand
        )
        if operands is None:
            operands = []
        self.set(state, operands)
        return bool(operands)

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        types = parser.parse_optional_undelimited_comma_separated_list(
            parser.parse_optional_type, parser.parse_type
        )
        ret = types is None
        if ret:
            types = ()
        self.set_types(state, types)
        return ret

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        state.operand_types[self.index] = (parser.parse_type(),)

    def get(self, op: IRDLOperation) -> VarOperand:
        return getattr(op, self.name)

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        return self.get(op).types

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        operand = self.get(op)
        if not operand:
            return
        state.print_whitespace(printer)
        printer.print_list(operand, printer.print_ssa_value)

    def set_empty(self, state: ParsingState):
        self.set(state, ())


class OptionalOperandVariable(OptionalVariable, OperandDirective):
    """
    An optional operand variable, with the following format:
      operand-directive ::= ( percent-ident )?
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, operand: UnresolvedOperand | None):
        state.operands[self.index] = () if operand is None else (operand,)

    def set_types(self, state: ParsingState, types: Sequence[Attribute]):
        state.operand_types[self.index] = types

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        operand = parser.parse_optional_unresolved_operand()
        self.set(state, operand)
        return bool(operand)

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        type = parser.parse_optional_type()
        self.set_types(state, () if type is None else (type,))
        return type is None

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        self.set_types(state, (parser.parse_type(),))

    def get(self, op: IRDLOperation) -> SSAValue | None:
        return getattr(op, self.name)

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        operand = self.get(op)
        if operand:
            return (operand.type,)
        return ()

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        operand = self.get(op)
        if not operand:
            return
        state.print_whitespace(printer)
        printer.print_ssa_value(operand)

    def set_empty(self, state: ParsingState):
        self.set(state, None)


_T = TypeVar("_T")


@dataclass(frozen=True)
class OperandsOrResultDirective(TypeableDirective, ABC):
    """
    Base class for the 'operands' and 'results' directives.
    """

    variadic_index: tuple[bool, int] | None
    """
    Represents the position of a (single) variadic variable, with the boolean
    representing whether it is optional
    """

    def is_variadic_like(self) -> bool:
        return True

    def is_anchorable(self) -> bool:
        return True

    def _set_using_variadic_index(
        self,
        field: list[None | Sequence[_T]],
        field_name: str,
        set_to: Sequence[_T],
    ) -> str | None:
        if self.variadic_index is None:
            if len(set_to) != len(field):
                return f"Expected {len(field)} {field_name} but found {len(set_to)}"
            field[:] = ((x,) for x in set_to)
            return

        is_optional, var_position = self.variadic_index
        var_length = len(set_to) - len(field) + 1
        if var_length < 0:
            return f"Expected at least {len(field) - 1} {field_name} but found {len(set_to)}"
        if var_length > 1 and is_optional:
            return f"Expected at most {len(field)} {field_name} but found {len(set_to)}"
        field[:var_position] = ((x,) for x in set_to[:var_position])
        field[var_position] = set_to[var_position : var_position + var_length]
        field[var_position + 1 :] = ((x,) for x in set_to[var_position + var_length :])


class OperandsDirective(OperandsOrResultDirective, FormatDirective):
    """
    An operands directive, with the following format:
      operands-directive ::= operands
    Prints each operand of the operation, inserting a comma between each.
    """

    def set_types(self, state: ParsingState, types: Sequence[Attribute]) -> str | None:
        return self._set_using_variadic_index(
            state.operand_types, "operand types", types
        )

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        pos_start = parser.pos
        operands = (
            parser.parse_optional_undelimited_comma_separated_list(
                parser.parse_optional_unresolved_operand,
                parser.parse_unresolved_operand,
            )
            or []
        )

        if s := self._set_using_variadic_index(state.operands, "operands", operands):
            parser.raise_error(s, at_position=pos_start, end_position=parser.pos)
        return bool(operands)

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        pos_start = parser.pos
        types = (
            parser.parse_optional_undelimited_comma_separated_list(
                parser.parse_optional_type, parser.parse_type
            )
            or []
        )

        if s := self.set_types(state, types):
            parser.raise_error(s, at_position=pos_start, end_position=parser.pos)
        return bool(types)

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        pos_start = parser.pos
        if s := self.set_types(state, (parser.parse_type(),)):
            parser.raise_error(s, at_position=pos_start, end_position=parser.pos)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        state.print_whitespace(printer)
        printer.print_list(op.operands, printer.print_ssa_value)

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        return op.operand_types

    def set_empty(self, state: ParsingState):
        state.operands = [() for _ in state.operands]

    def is_present(self, op: IRDLOperation) -> bool:
        return bool(op.operands)


@dataclass(frozen=True)
class ResultVariable(VariableDirective, TypeableDirective):
    """
    An result variable, with the following format:
      result-directive ::= dollar-ident
    This directive can not be used for parsing and printing directly, as result
    parsing is not handled by the custom operation parser.
    """

    def set_types(self, state: ParsingState, types: Sequence[Attribute]):
        state.result_types[self.index] = types

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        self.set_types(state, (parser.parse_type(),))
        return True

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        self.parse_types(parser, state)

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        return (getattr(op, self.name).type,)


@dataclass(frozen=True)
class VariadicResultVariable(VariadicVariable, TypeableDirective):
    """
    A variadic result variable, with the following format:
      result-directive ::= percent-ident (( `,` percent-id )* )?
    This directive can not be used for parsing and printing directly, as result
    parsing is not handled by the custom operation parser.
    """

    def set_types(self, state: ParsingState, types: Sequence[Attribute]):
        state.result_types[self.index] = types

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        types = parser.parse_optional_undelimited_comma_separated_list(
            parser.parse_optional_type, parser.parse_type
        )
        self.set_types(state, () if types is None else types)
        return types is None

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        state.result_types[self.index] = (parser.parse_type(),)

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        return getattr(op, self.name).types


class OptionalResultVariable(OptionalVariable, TypeableDirective):
    """
    An optional result variable, with the following format:
      result-directive ::= ( percent-ident )?
    This directive can not be used for parsing and printing directly, as result
    parsing is not handled by the custom operation parser.
    """

    def set_types(self, state: ParsingState, types: Sequence[Attribute]):
        state.result_types[self.index] = types

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        type = parser.parse_optional_type()
        self.set_types(state, () if type is None else (type,))
        return type is not None

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        self.set_types(state, (parser.parse_type(),))

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        res = getattr(op, self.name)
        if res:
            return (res.type,)
        return ()


class ResultsDirective(OperandsOrResultDirective):
    """
    A results directive, with the following format:
      results-directive ::= results
    A typeable directive which processes the result types of the operation.
    """

    def set_types(self, state: ParsingState, types: Sequence[Attribute]) -> str | None:
        return self._set_using_variadic_index(state.result_types, "result types", types)

    def parse_types(self, parser: Parser, state: ParsingState) -> bool:
        pos_start = parser.pos
        types = (
            parser.parse_optional_undelimited_comma_separated_list(
                parser.parse_optional_type, parser.parse_type
            )
            or []
        )

        if s := self.set_types(state, types):
            parser.raise_error(s, at_position=pos_start, end_position=parser.pos)
        return bool(types)

    def parse_single_type(self, parser: Parser, state: ParsingState) -> None:
        pos_start = parser.pos
        if s := self.set_types(state, (parser.parse_type(),)):
            parser.raise_error(s, at_position=pos_start, end_position=parser.pos)

    def get_types(self, op: IRDLOperation) -> Sequence[Attribute]:
        return op.result_types

    def is_present(self, op: IRDLOperation) -> bool:
        return bool(op.results)


@dataclass(frozen=True)
class FunctionalTypeDirective(FormatDirective):
    """
    A directive which parses a functional type, with format:
      functional-type-directive ::= functional-type(typeable-directive, typeable-directive)
    A functional type is either of the form
      `(` type-list `)` `->` `(` type-list `)`
    or
      `(` type-list `)` `->` type
    where type-list is a comma separated list of types (or the empty string to signify the empty list).
    The second format is preferred for printing when possible.
    """

    operand_typeable_directive: TypeableDirective
    result_typeable_directive: TypeableDirective

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        if not parser.parse_optional_punctuation("("):
            return False
        self.operand_typeable_directive.parse_types(parser, state)
        parser.parse_punctuation(")")
        parser.parse_punctuation("->")
        if parser.parse_optional_punctuation("("):
            self.result_typeable_directive.parse_types(parser, state)
            parser.parse_punctuation(")")
        else:
            self.result_typeable_directive.parse_single_type(parser, state)
        return True

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        state.print_whitespace(printer)
        with printer.in_parens():
            printer.print_list(
                self.operand_typeable_directive.get_types(op), printer.print_attribute
            )
        printer.print_string(" -> ")
        result_types = self.result_typeable_directive.get_types(op)
        if len(result_types) == 1:
            printer.print_attribute(result_types[0])
        else:
            with printer.in_parens():
                printer.print_list(result_types, printer.print_attribute)


class RegionDirective(FormatDirective, ABC):
    """
    Baseclass to help keep typechecking simple.
    """


@dataclass(frozen=True)
class RegionVariable(RegionDirective, VariableDirective):
    """
    A region variable, with the following format:
      region-directive ::= dollar-ident
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, region: Region):
        state.regions[self.index] = (region,)

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        self.set(state, parser.parse_region())
        return True

    def get(self, op: IRDLOperation) -> Region:
        return getattr(op, self.name)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        state.print_whitespace(printer)
        printer.print_region(self.get(op))

    def is_anchorable(self) -> bool:
        return True

    def set_empty(self, state: ParsingState):
        self.set(state, Region())

    def is_present(self, op: IRDLOperation) -> bool:
        return bool(self.get(op).blocks)


@dataclass(frozen=True)
class VariadicRegionVariable(RegionDirective, VariadicVariable):
    """
    A variadic region variable, with the following format:
      region-directive ::= dollar-ident

    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, region: Sequence[Region]):
        state.regions[self.index] = region

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        regions: list[Region] = []
        current_region = parser.parse_optional_region()
        while current_region is not None:
            regions.append(current_region)
            current_region = parser.parse_optional_region()

        self.set(state, regions)
        return bool(regions)

    def get(self, op: IRDLOperation) -> Sequence[Region]:
        return getattr(op, self.name)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        region = self.get(op)
        if not region:
            return
        state.print_whitespace(printer)
        printer.print_list(region, printer.print_region, delimiter=" ")

    def set_empty(self, state: ParsingState):
        self.set(state, ())


class OptionalRegionVariable(RegionDirective, OptionalVariable):
    """
    An optional region variable, with the following format:
      region-directive ::= dollar-ident
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, region: Region | None):
        state.regions[self.index] = () if region is None else (region,)

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        region = parser.parse_optional_region()
        self.set(state, region)
        return region is not None

    def get(self, op: IRDLOperation) -> Region | None:
        return getattr(op, self.name)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        region = self.get(op)
        if not region:
            return
        state.print_whitespace(printer)
        printer.print_region(region)

    def set_empty(self, state: ParsingState):
        self.set(state, None)


class SuccessorDirective(FormatDirective, ABC):
    """
    Base class for type checking.
    A variadic successor directive cannot follow another variadic successor directive.
    """


class SuccessorVariable(VariableDirective, SuccessorDirective):
    """
    A successor variable, with the following format:
      successor-directive ::= dollar-ident
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, successor: Successor):
        state.successors[self.index] = (successor,)

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        successor = parser.parse_successor()

        self.set(state, successor)

        return True

    def get(self, op: IRDLOperation) -> Successor:
        return getattr(op, self.name)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        state.print_whitespace(printer)
        printer.print_block_name(self.get(op))


class VariadicSuccessorVariable(VariadicVariable, SuccessorDirective):
    """
    A variadic successor variable, with the following format:
      successor-directive ::= dollar-ident
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, successors: Sequence[Successor]):
        state.successors[self.index] = successors

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        successors = (
            parser.parse_optional_undelimited_comma_separated_list(
                parser.parse_optional_successor, parser.parse_successor
            )
            or []
        )

        self.set(state, successors)

        return bool(successors)

    def get(self, op: IRDLOperation) -> Sequence[Successor]:
        return getattr(op, self.name)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        successor = self.get(op)
        if not successor:
            return
        state.print_whitespace(printer)
        printer.print_list(successor, printer.print_block_name)

    def set_empty(self, state: ParsingState):
        self.set(state, ())


class OptionalSuccessorVariable(OptionalVariable, SuccessorDirective):
    """
    An optional successor variable, with the following format:
      successor-directive ::= dollar-ident
    The directive will request a space to be printed after.
    """

    def set(self, state: ParsingState, successor: Successor | None):
        state.successors[self.index] = () if successor is None else (successor,)

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        successor = parser.parse_optional_successor()
        self.set(state, successor)
        return successor is not None

    def get(self, op: IRDLOperation) -> Successor | None:
        return getattr(op, self.name)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        successor = self.get(op)
        if not successor:
            return
        state.print_whitespace(printer)
        printer.print_block_name(successor)

    def set_empty(self, state: ParsingState):
        self.set(state, None)


@dataclass(frozen=True)
class AttributeVariable(FormatDirective):
    """
    An attribute variable, with the following format:
      attribute-variable ::= dollar-ident
    The directive will request a space to be printed right after.
    """

    name: str
    """The attribute name as it should be in the attribute or property dictionary."""
    is_property: bool
    """Should this attribute be put in the attribute or property dictionary."""
    is_optional: bool
    """Is this attribute optional in the operation definition."""
    default_value: Attribute | None

    def set(self, state: ParsingState, attr: Attribute):
        if self.is_property:
            state.properties[self.name] = attr
        else:
            state.attributes[self.name] = attr

    def parse_attr(self, parser: Parser) -> Attribute | None:
        if self.is_optional:
            return parser.parse_optional_attribute()
        else:
            return parser.parse_attribute()

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        attr = self.parse_attr(parser)
        if attr is None:
            return False
        self.set(state, attr)
        return True

    def get(self, op: IRDLOperation) -> Attribute | None:
        if self.is_property:
            return op.properties.get(self.name)
        else:
            return op.attributes.get(self.name)

    def print_attr(self, printer: Printer, attr: Attribute) -> None:
        return printer.print_attribute(attr)

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        attr = self.get(op)

        if attr is None or attr == self.default_value:
            return

        state.print_whitespace(printer)

        self.print_attr(printer, attr)

    def is_present(self, op: IRDLOperation) -> bool:
        attr = self.get(op)
        return attr is not None and attr != self.default_value

    def is_anchorable(self) -> bool:
        return self.is_optional or self.default_value is not None

    def is_optional_like(self) -> bool:
        return self.is_optional


@dataclass(frozen=True)
class UniqueBaseAttributeVariable(AttributeVariable):
    unique_base: type[Attribute]
    """The known base class of the Attribute, if any."""

    def parse_attr(self, parser: Parser) -> Attribute | None:
        unique_base = self.unique_base

        if issubclass(unique_base, ParametrizedAttribute):
            return unique_base.new(unique_base.parse_parameters(parser))
        elif issubclass(unique_base, Data):
            unique_base = cast(type[Data[Any]], unique_base)
            return unique_base.new(unique_base.parse_parameter(parser))
        else:
            raise ValueError("Attributes must be Data or ParameterizedAttribute.")

    def print_attr(self, printer: Printer, attr: Attribute) -> None:
        if isinstance(attr, ParametrizedAttribute):
            return attr.print_parameters(printer)
        if isinstance(attr, Data):
            return attr.print_parameter(printer)
        raise ValueError("Attributes must be Data or ParameterizedAttribute!")


@dataclass(frozen=True)
class TypedAttributeVariable(UniqueBaseAttributeVariable):
    unique_type: Attribute
    """The known type of the Attribute, if any."""

    def parse_attr(self, parser: Parser) -> Attribute | None:
        unique_base = self.unique_base
        assert issubclass(unique_base, TypedAttribute)
        return unique_base.parse_with_type(parser, self.unique_type)

    def print_attr(self, printer: Printer, attr: Attribute) -> None:
        assert isinstance(attr, TypedAttribute)
        return attr.print_without_type(printer)


@dataclass(frozen=True)
class DenseArrayAttributeVariable(AttributeVariable):
    elt_type: IntegerType | AnyFloat

    def parse_attr(self, parser: Parser) -> Attribute | None:
        if isinstance(self.elt_type, IntegerType):
            if self.is_optional:
                elements = parser.parse_optional_comma_separated_list(
                    parser.Delimiter.SQUARE, parser.parse_integer
                )
                if elements is None:
                    return None
            else:
                elements = parser.parse_comma_separated_list(
                    parser.Delimiter.SQUARE, parser.parse_integer
                )
            return DenseArrayBase(
                self.elt_type, BytesAttr(self.elt_type.pack(elements))
            )
        else:
            if self.is_optional:
                elements = parser.parse_optional_comma_separated_list(
                    parser.Delimiter.SQUARE, parser.parse_float
                )
                if elements is None:
                    return None
            else:
                elements = parser.parse_comma_separated_list(
                    parser.Delimiter.SQUARE, parser.parse_float
                )
            return DenseArrayBase(
                self.elt_type, BytesAttr(self.elt_type.pack(elements))
            )

    def print_attr(self, printer: Printer, attr: Attribute) -> None:
        with printer.in_square_brackets():
            if isa(attr, DenseArrayBase[IntegerType]):
                printer.print_list(attr.iter_values(), printer.print_int)
            elif isa(attr, DenseArrayBase[AnyFloat]):
                printer.print_list(
                    attr.iter_values(),
                    lambda value: printer.print_float(value, attr.elt_type),
                )


class SymbolNameAttributeVariable(AttributeVariable):
    def parse_attr(self, parser: Parser) -> Attribute | None:
        if self.is_optional:
            return parser.parse_optional_symbol_name()
        else:
            return parser.parse_symbol_name()

    def print_attr(self, printer: Printer, attr: Attribute) -> None:
        assert isinstance(attr, StringAttr)
        return printer.print_symbol_name(attr.data)


class OptionalUnitAttrVariable(AttributeVariable):
    """
    An optional UnitAttr variable that holds no value and derives its meaning from its existence. Holds a parse
    and print method to reflect this.

      operand-directive ::= (`unit_attr` unit_attr^)?

    Also see: https://mlir.llvm.org/docs/DefiningDialects/Operations/#unit-attributes
    """

    def __init__(self, name: str, is_property: bool):
        super().__init__(name, is_property, True, None)

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        self.set(state, UnitAttr())
        return True

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        return

    def is_optional_like(self) -> bool:
        return False


@dataclass(frozen=True)
class WhitespaceDirective(FormatDirective):
    """
    A whitespace directive, with the following format:
      whitespace-directive ::= `\n` | ` ` | ``
    This directive is only applied during printing, and has no effect during
    parsing.
    The directive will not request any space to be printed after.
    """

    whitespace: Literal[" ", "\n", ""]
    """The whitespace that should be printed."""

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        return False

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        printer.print_string(self.whitespace)
        state.last_was_punctuation = self.whitespace == ""
        state.should_emit_space = False


@dataclass(frozen=True)
class PunctuationDirective(FormatDirective):
    """
    A punctuation directive, with the following format:
      punctuation-directive ::= punctuation
    The directive will request a space to be printed right after, unless the punctuation
    is `<`, `(`, `{`, or `[`.
    It will also print a space before if a space is requested, and that the punctuation
    is neither `>`, `)`, `}`, `]`, or `,` if the last element was a punctuation, and
    additionally neither `<`, `(`, `}`, `]`, if the last element was not a punctuation.
    """

    punctuation: PunctuationSpelling
    """The punctuation that should be printed/parsed."""

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        return parser.parse_optional_punctuation(self.punctuation) is not None

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        emit_space = False
        if state.should_emit_space:
            if state.last_was_punctuation:
                if self.punctuation not in (">", ")", "}", "]", ","):
                    emit_space = True
            elif self.punctuation not in ("<", ">", "(", ")", "{", "}", "[", "]", ","):
                emit_space = True

            if emit_space:
                printer.print_string(" ")

        printer.print_string(self.punctuation)

        state.should_emit_space = self.punctuation not in ("<", "(", "{", "[")
        state.last_was_punctuation = True

    def is_optional_like(self) -> bool:
        return True


@dataclass(frozen=True)
class KeywordDirective(FormatDirective):
    """
    A keyword directive, with the following format:
      keyword-directive ::= bare-ident
    The directive expects a specific identifier, and will request a space to be printed
    after.
    """

    keyword: str
    """The identifier that should be printed."""

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        return parser.parse_optional_keyword(self.keyword) is not None

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        if state.should_emit_space:
            printer.print_string(" ")
        state.should_emit_space = True
        state.last_was_punctuation = False

        printer.print_string(self.keyword)

    def is_optional_like(self) -> bool:
        return True


@dataclass(frozen=True)
class OptionalGroupDirective(FormatDirective):
    anchor: Directive
    then_whitespace: tuple[WhitespaceDirective, ...]
    then_first: FormatDirective
    then_elements: tuple[FormatDirective, ...]
    else_elements: tuple[FormatDirective, ...]

    def parse(self, parser: Parser, state: ParsingState) -> bool:
        # If the first element was parsed, parse the then-elements as usual
        if ret := self.then_first.parse(parser, state):
            for element in self.then_elements:
                element.parse(parser, state)
            for element in self.else_elements:
                element.set_empty(state)
        else:
            for element in self.then_elements:
                element.set_empty(state)
            for element in self.else_elements:
                element.parse(parser, state)
        return ret

    def print(self, printer: Printer, state: PrintingState, op: IRDLOperation) -> None:
        if self.anchor.is_present(op):
            for element in (
                *self.then_whitespace,
                self.then_first,
                *self.then_elements,
            ):
                element.print(printer, state, op)
        else:
            for element in self.else_elements:
                element.print(printer, state, op)

    def set_empty(self, state: ParsingState) -> None:
        self.then_first.set_empty(state)
        for element in self.then_elements:
            element.set_empty(state)
        for element in self.else_elements:
            element.set_empty(state)



#!/usr/bin/env python3
"""Benchmarks for the verifier of the xDSL implementation."""

from benchmarks.workloads import WorkloadBuilder


class Verifier:
    """Benchmark verification in xDSL.

    For a single rewriting pass, we verify with the input before the pass and
    the output after the pass.

    Note that this is run on the parsed input, rather than the output of the
    re-writing pass. This is because constant folding and dead-code elimination
    reduce the rewritten results to negligibly small sizes, so only the first
    verification pass contributes to the overall performance.
    """

    WORKLOAD_CONSTANT_100 = WorkloadBuilder.constant_folding_module(100)
    WORKLOAD_CONSTANT_1000 = WorkloadBuilder.constant_folding_module(1000)
    WORKLOAD_LARGE_DENSE_ATTR = WorkloadBuilder.large_dense_attr_module()

    def time_constant_100(self) -> None:
        """Time verifying constant folding for 100 items."""
        Verifier.WORKLOAD_CONSTANT_100.verify()

    def time_constant_1000(self) -> None:
        """Time verifying constant folding for 1000 items."""
        Verifier.WORKLOAD_CONSTANT_1000.verify()

    def time_dense_attr_hex(self) -> None:
        """Time verifying a 1024x1024xi8 dense attribute given as a hex string."""
        Verifier.WORKLOAD_LARGE_DENSE_ATTR.verify()


if __name__ == "__main__":
    from bench_utils import Benchmark, profile

    VERIFIER = Verifier()
    profile(
        {
            "Verifier.constant_100": Benchmark(VERIFIER.time_constant_100),
            "Verifier.constant_1000": Benchmark(VERIFIER.time_constant_1000),
            "Verifier.dense_attr_hex": Benchmark(VERIFIER.time_dense_attr_hex),
        }
    )



from xdsl.builder import Builder
from xdsl.dialects.builtin import IntegerAttr, IntegerType, MemRefType, TensorType, f32
from xdsl.dialects.csl.csl_stencil import AccessOp, ApplyOp
from xdsl.dialects.stencil import IndexAttr, TempType
from xdsl.ir import Region, SSAValue
from xdsl.utils.test_value import create_ssa_value


def test_access_patterns():
    temp_t = TempType(5, f32)
    temp = create_ssa_value(temp_t)
    mref = create_ssa_value(mref_t := MemRefType(tens_t := TensorType(f32, (5,)), (4,)))

    @Builder.implicit_region((mref_t, temp_t))
    def region0(args: tuple[SSAValue, ...]):
        t0, t1 = args
        for x in (-1, 1):
            AccessOp(t0, IndexAttr.get(x, 0), tens_t)
        for y in (-1, 1):
            AccessOp(t0, IndexAttr.get(0, y), tens_t)

        AccessOp(t1, IndexAttr.get(1, 1), tens_t)
        AccessOp(t1, IndexAttr.get(-1, -1), tens_t)

    @Builder.implicit_region((temp_t, temp_t))
    def region1(args: tuple[SSAValue, ...]):
        t0, t1 = args
        AccessOp(t0, IndexAttr.get(0, 0), tens_t)
        AccessOp(t1, IndexAttr.get(0, 0), tens_t)

    apply = ApplyOp(
        operands=[temp, mref, [], [], []],
        properties={
            "swaps": None,
            "topo": None,
            "num_chunks": IntegerAttr(1, IntegerType(64)),
        },
        regions=[
            Region(region0.detach_block(0)),
            Region(region1.detach_block(0)),
        ],
        result_types=[tens_t],
    )

    r0_t0_acc, r0_t1_acc, r1_t0_acc, r1_t1_acc = tuple(apply.get_accesses())

    assert r0_t0_acc.visual_pattern() == " X \nXOX\n X "
    assert r0_t1_acc.visual_pattern() == "X  \n O \n  X"

    assert not r0_t0_acc.is_diagonal
    assert r0_t1_acc.is_diagonal

    assert len(tuple(r0_t1_acc.get_diagonals())) == 2

    assert r1_t0_acc.visual_pattern() == "X"
    assert r1_t1_acc.visual_pattern() == "X"
